"""FMP Discovery Filings Model."""

# pylint: disable=unused-argument

from datetime import datetime, timedelta
from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.discovery_filings import (
    DiscoveryFilingsData,
    DiscoveryFilingsQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import Field, model_validator


class FMPDiscoveryFilingsQueryParams(DiscoveryFilingsQueryParams):
    """FMP Discovery Filings Query.

    Source: https://site.financialmodelingprep.com/developer/docs#search-by-form-type
    """

    __alias_dict__ = {
        "start_date": "from",
        "end_date": "to",
        "form_type": "formType",
    }

    limit: int | None = Field(
        default=None,
        description="The maximum number of results to return. Default is 10000.",
    )

    @model_validator(mode="before")
    @classmethod
    def _check_date_range(cls, values):
        """Validate date range."""
        start_date = values.get("start_date")
        end_date = values.get("end_date")

        # Validate date range
        if start_date and end_date and end_date - start_date > timedelta(days=90):
            raise ValueError("Date range cannot exceed 90 days.")

        return values


class FMPDiscoveryFilingsData(DiscoveryFilingsData):
    """FMP Discovery Filings Data."""

    final_link: str = Field(
        description="Direct URL to the main document of the filing."
    )


class FMPDiscoveryFilingsFetcher(
    Fetcher[
        FMPDiscoveryFilingsQueryParams,
        list[FMPDiscoveryFilingsData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPDiscoveryFilingsQueryParams:
        """Transform the query."""
        return FMPDiscoveryFilingsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPDiscoveryFilingsQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import math  # noqa
        from openbb_core.provider.utils.helpers import amake_requests, get_querystring

        api_key = credentials.get("fmp_api_key") if credentials else ""
        data: list[dict] = []
        limit = query.limit or 10000
        base_url = (
            "https://financialmodelingprep.com/stable/sec-filings-search/form-type"
            if query.form_type
            else "https://financialmodelingprep.com/stable/sec-filings-financials/"
        )
        start_date = (
            query.start_date
            or (datetime.now() - timedelta(days=89 if query.form_type else 2)).date()
        )
        end_date = query.end_date or datetime.now().date()
        query.start_date = start_date
        query.end_date = end_date

        query_str = get_querystring(query.model_dump(by_alias=True), ["limit"])

        # FMP only allows 1000 results per page
        pages = math.ceil(limit / 1000)

        urls = [
            f"{base_url}?{query_str}&page={page}&limit=1000&apikey={api_key}"
            for page in range(pages)
        ]

        data = await amake_requests(urls, **kwargs)

        return sorted(data, key=lambda x: x["acceptedDate"], reverse=True)

    @staticmethod
    def transform_data(
        query: FMPDiscoveryFilingsQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPDiscoveryFilingsData]:
        """Return the transformed data."""
        if not data:
            raise EmptyDataError("No data was returned for the given query.")
        return [FMPDiscoveryFilingsData.model_validate(d) for d in data]



"""FMP Insider Trading Model."""

# pylint: disable=unused-argument

from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.insider_trading import (
    InsiderTradingData,
    InsiderTradingQueryParams,
)
from openbb_fmp.utils.definitions import TRANSACTION_TYPES, TRANSACTION_TYPES_DICT
from pydantic import Field, field_validator


class FMPInsiderTradingQueryParams(InsiderTradingQueryParams):
    """FMP Insider Trading Query.

    Source: https://site.financialmodelingprep.com/developer/docs#search-insider-trades
    """

    __alias_dict__ = {
        "transaction_type": "transactionType",
    }
    __json_schema_extra__ = {
        "transaction_type": {
            "x-widget_config": {
                "options": [
                    {"label": v, "value": k} for k, v in TRANSACTION_TYPES_DICT.items()
                ],
            }
        }
    }

    transaction_type: TRANSACTION_TYPES | None = Field(
        default=None,
        description="Type of the transaction.",
    )

    statistics: bool = Field(
        default=False,
        description="Flag to return summary statistics for the given symbol."
        + " Setting as True will ignore other parameters except symbol.",
    )

    @field_validator("transaction_type", mode="before", check_fields=False)
    @classmethod
    def validate_transaction_type(cls, v):
        """Validate the transaction type."""
        if isinstance(v, list):
            return ",".join(v)
        return v


class FMPInsiderTradingData(InsiderTradingData):
    """FMP Insider Trading Data."""

    __alias_dict__ = {
        "owner_cik": "reportingCik",
        "owner_name": "reportingName",
        "owner_title": "typeOfOwner",
        "ownership_type": "directOrIndirect",
        "security_type": "securityName",
        "transaction_price": "price",
        "acquisition_or_disposition": "acquisitionOrDisposition",
        "filing_url": "link",
        "transactions_ratio": "acquiredDisposedRatio",
        "company_cik": "cik",
    }

    form_type: str | None = Field(default=None, description="The SEC form type.")
    year: int | None = Field(
        default=None, description="The calendar year for the statistics."
    )
    quarter: int | None = Field(
        default=None, description="The calendar quarter for the statistics."
    )
    acquired_transactions: int | None = Field(
        default=None, description="Number of acquired transactions (statistics only)."
    )
    disposed_transactions: int | None = Field(
        default=None, description="Number of disposed transactions (statistics only)."
    )
    transactions_ratio: float | None = Field(
        default=None,
        description="Ratio of acquired to disposed transactions (statistics only).",
    )
    total_acquired: int | float | None = Field(
        default=None, description="Total number of shares acquired (statistics only)."
    )
    total_disposed: int | float | None = Field(
        default=None, description="Total number of shares disposed (statistics only)."
    )
    average_acquired: float | None = Field(
        default=None,
        description="Average number of shares acquired per transaction (statistics only).",
    )
    average_disposed: float | None = Field(
        default=None,
        description="Average number of shares disposed per transaction (statistics only).",
    )
    total_purchases: int | None = Field(
        default=None,
        description="Total number of purchase transactions (statistics only).",
    )
    total_sales: int | None = Field(
        default=None, description="Total number of sale transactions (statistics only)."
    )


class FMPInsiderTradingFetcher(
    Fetcher[
        FMPInsiderTradingQueryParams,
        list[FMPInsiderTradingData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPInsiderTradingQueryParams:
        """Transform the query params."""
        return FMPInsiderTradingQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPInsiderTradingQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import math  # noqa
        from openbb_core.provider.utils.helpers import get_querystring
        from openbb_fmp.utils.helpers import get_data_urls, get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""

        if query.statistics is True:
            url = (
                "https://financialmodelingprep.com/stable/insider-trading/statistics"
                + f"?symbol={query.symbol}&apikey={api_key}"
            )
            return await get_data_many(url, **kwargs)

        transaction_type = (
            TRANSACTION_TYPES_DICT.get(query.transaction_type)
            if query.transaction_type
            else None
        )
        limit = query.limit if query.limit and query.limit <= 1000 else 1000
        query = query.model_copy(update={"transaction_type": transaction_type})
        base_url = "https://financialmodelingprep.com/stable/insider-trading/search"
        query_str = get_querystring(query.model_dump(by_alias=True), ["page", "limit"])

        pages = math.ceil(limit / 1000)
        urls = [
            f"{base_url}?{query_str}&page={page}&limit={limit}&apikey={api_key}"
            for page in range(pages)
        ]

        return await get_data_urls(urls, **kwargs)  # type: ignore

    @staticmethod
    def transform_data(
        query: FMPInsiderTradingQueryParams, data: list, **kwargs: Any
    ) -> list[FMPInsiderTradingData]:
        """Return the transformed data."""
        return (
            [
                FMPInsiderTradingData.model_validate(d)
                for d in sorted(data, key=lambda x: x["filingDate"], reverse=True)
            ]
            if query.statistics is False
            else [
                FMPInsiderTradingData.model_validate(d)
                for d in sorted(
                    data, key=lambda x: (x["year"], x["quarter"]), reverse=True
                )
            ]
        )



"""FMP Historical Dividends Model."""

# pylint: disable=unused-argument

from datetime import date as dateType
from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.historical_dividends import (
    HistoricalDividendsData,
    HistoricalDividendsQueryParams,
)
from pydantic import Field, field_validator


class FMPHistoricalDividendsQueryParams(HistoricalDividendsQueryParams):
    """FMP Historical Dividends Query.

    Source: https://site.financialmodelingprep.com/developer/docs#dividends-company
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}

    limit: int | None = Field(
        default=None, description="Return N most recent payments."
    )


class FMPHistoricalDividendsData(HistoricalDividendsData):
    """FMP Historical Dividends Data."""

    __alias_dict__ = {
        "ex_dividend_date": "date",
        "amount": "dividend",
        "adjusted_amount": "adjDividend",
        "dividend_yield": "yield",
        "record_date": "recordDate",
        "payment_date": "paymentDate",
        "declaration_date": "declarationDate",
    }
    declaration_date: dateType | None = Field(
        default=None,
        description="Declaration date of the historical dividends.",
    )
    record_date: dateType | None = Field(
        default=None,
        description="Record date of the historical dividends.",
    )
    payment_date: dateType | None = Field(
        default=None,
        description="Payment date of the historical dividends.",
    )
    adjusted_amount: float = Field(description="Split-adjusted dividend amount.")
    dividend_yield: float | None = Field(
        default=None,
        description="Dividend yield represented by the payment.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    frequency: str | None = Field(default=None, description="Frequency of the payment.")

    @field_validator(
        "dividend_yield",
        mode="before",
        check_fields=False,
    )
    @classmethod
    def _normalize_percent(cls, v):
        """Validate dates."""
        return v / 100 if v else None


class FMPHistoricalDividendsFetcher(
    Fetcher[
        FMPHistoricalDividendsQueryParams,
        list[FMPHistoricalDividendsData],
    ]
):
    """FMP Historical Dividend Yield Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPHistoricalDividendsQueryParams:
        """Transform the query params."""
        return FMPHistoricalDividendsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPHistoricalDividendsQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import warnings  # noqa
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        limit = query.limit or 1000
        symbols = query.symbol.split(",")  # type: ignore
        results: list = []

        for symbol in symbols:
            url = f"https://financialmodelingprep.com/stable/dividends?symbol={symbol}&limit={limit}&apikey={api_key}"
            result = await get_data_many(url, "historical", **kwargs)

            if not result:
                warnings.warn(f"No data found for symbol {symbol}")
                continue

            results.extend(result)

        return sorted(results, key=lambda x: x.get("date"), reverse=True)

    @staticmethod
    def transform_data(
        query: FMPHistoricalDividendsQueryParams, data: list, **kwargs: Any
    ) -> list[FMPHistoricalDividendsData]:
        """Return the transformed data."""
        result: list[FMPHistoricalDividendsData] = []

        for d in data:
            d["declarationDate"] = d.get("declarationDate") or None

            if query.start_date or query.end_date:
                dt = d.get("date")

                if not dt:
                    continue

                if query.start_date and dt < query.start_date.isoformat():
                    continue

                if query.end_date and dt > query.end_date.isoformat():
                    continue

            result.append(FMPHistoricalDividendsData(**d))

        return result



"""FMP Income Statement Model."""

# pylint: disable=unused-argument

from datetime import (
    date as dateType,
    datetime,
)
from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.income_statement import (
    IncomeStatementData,
    IncomeStatementQueryParams,
)
from openbb_core.provider.utils.descriptions import QUERY_DESCRIPTIONS
from openbb_fmp.utils.definitions import FinancialStatementPeriods
from pydantic import Field


class FMPIncomeStatementQueryParams(IncomeStatementQueryParams):
    """FMP Income Statement Query.

    Source: https://financialmodelingprep.com/developer/docs/#Income-Statement
    """

    period: FinancialStatementPeriods = Field(
        default="annual",
        description=QUERY_DESCRIPTIONS.get("period", ""),
    )


class FMPIncomeStatementData(IncomeStatementData):
    """FMP Income Statement Data."""

    __alias_dict__ = {
        "period_ending": "date",
        "fiscal_period": "period",
        "fiscal_year": "calendarYear",
        "filing_date": "fillingDate",
        "accepted_date": "acceptedDate",
        "reported_currency": "reportedCurrency",
        "revenue": "revenue",
        "cost_of_revenue": "costOfRevenue",
        "gross_profit": "grossProfit",
        "general_and_admin_expense": "generalAndAdministrativeExpenses",
        "research_and_development_expense": "researchAndDevelopmentExpenses",
        "selling_and_marketing_expense": "sellingAndMarketingExpenses",
        "selling_general_and_admin_expense": "sellingGeneralAndAdministrativeExpenses",
        "other_expenses": "otherExpenses",
        "total_operating_expenses": "operatingExpenses",
        "cost_and_expenses": "costAndExpenses",
        "interest_income": "interestIncome",
        "total_interest_expense": "interestExpense",
        "depreciation_and_amortization": "depreciationAndAmortization",
        "ebitda": "ebitda",
        "total_operating_income": "operatingIncome",
        "total_other_income_expenses": "totalOtherIncomeExpensesNet",
        "total_pre_tax_income": "incomeBeforeTax",
        "income_tax_expense": "incomeTaxExpense",
        "consolidated_net_income": "netIncome",
        "basic_earnings_per_share": "eps",
        "diluted_earnings_per_share": "epsDiluted",
        "weighted_average_basic_shares_outstanding": "weightedAverageShsOut",
        "weighted_average_diluted_shares_outstanding": "weightedAverageShsOutDil",
    }

    filing_date: dateType | None = Field(
        default=None,
        description="The date when the filing was made.",
    )
    accepted_date: datetime | None = Field(
        default=None,
        description="The date and time when the filing was accepted.",
    )
    cik: str | None = Field(
        default=None,
        description="The Central Index Key (CIK) assigned by the SEC, if applicable.",
    )
    symbol: str | None = Field(
        default=None,
        description="The stock ticker symbol.",
    )
    reported_currency: str | None = Field(
        default=None,
        description="The currency in which the balance sheet was reported.",
    )
    revenue: int | None = Field(
        default=None,
        description="Total revenue.",
    )
    cost_of_revenue: int | None = Field(
        default=None,
        description="Cost of revenue.",
    )
    gross_profit: int | None = Field(
        default=None,
        description="Gross profit.",
    )
    general_and_admin_expense: int | None = Field(
        default=None,
        description="General and administrative expenses.",
    )
    research_and_development_expense: int | None = Field(
        default=None,
        description="Research and development expenses.",
    )
    selling_and_marketing_expense: int | None = Field(
        default=None,
        description="Selling and marketing expenses.",
    )
    selling_general_and_admin_expense: int | None = Field(
        default=None,
        description="Selling, general and administrative expenses.",
    )
    other_expenses: int | None = Field(
        default=None,
        description="Other expenses.",
    )
    total_operating_expenses: int | None = Field(
        default=None,
        description="Total operating expenses.",
    )
    cost_and_expenses: int | None = Field(
        default=None,
        description="Cost and expenses.",
    )
    interest_income: int | None = Field(
        default=None,
        description="Interest income.",
    )
    total_interest_expense: int | None = Field(
        default=None,
        description="Total interest expenses.",
    )
    net_interest_income: int | None = Field(
        default=None,
        description="Net interest income.",
    )
    depreciation_and_amortization: int | None = Field(
        default=None,
        description="Depreciation and amortization.",
    )
    ebit: int | None = Field(
        default=None,
        description="Earnings before interest and taxes (EBIT).",
        title="EBIT",
    )
    ebitda: int | None = Field(
        default=None,
        description="EBITDA.",
        title="EBITDA",
    )
    total_operating_income: int | None = Field(
        default=None,
        description="Total operating income.",
    )
    non_operating_income_excluding_interest: int | None = Field(
        default=None,
        description="Non-operating income excluding interest.",
    )
    net_income_from_continuing_operations: int | None = Field(
        default=None,
        description="Net income from continuing operations.",
    )
    net_income_from_discontinued_operations: int | None = Field(
        default=None,
        description="Net income from discontinued operations.",
    )
    total_other_income_expenses: int | None = Field(
        default=None,
        description="Total other income and expenses.",
    )
    total_pre_tax_income: int | None = Field(
        default=None,
        description="Total pre-tax income.",
    )
    income_tax_expense: int | None = Field(
        default=None,
        description="Income tax expense.",
    )
    other_adjustments_to_net_income: int | None = Field(
        default=None,
        description="Other adjustments to net income.",
    )
    net_income_deductions: int | None = Field(
        default=None,
        description="Net income deductions.",
    )
    consolidated_net_income: int | None = Field(
        default=None,
        description="Consolidated net income.",
    )
    bottom_line_net_income: int | None = Field(
        default=None,
        description="Bottom line net income.",
    )
    basic_earnings_per_share: float | None = Field(
        default=None,
        description="Basic earnings per share.",
    )
    diluted_earnings_per_share: float | None = Field(
        default=None,
        description="Diluted earnings per share.",
    )
    weighted_average_basic_shares_outstanding: int | None = Field(
        default=None,
        description="Weighted average basic shares outstanding.",
    )
    weighted_average_diluted_shares_outstanding: int | None = Field(
        default=None,
        description="Weighted average diluted shares outstanding.",
    )


class FMPIncomeStatementFetcher(
    Fetcher[
        FMPIncomeStatementQueryParams,
        list[FMPIncomeStatementData],
    ]
):
    """FMP Income Statement Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPIncomeStatementQueryParams:
        """Transform the query params."""
        return FMPIncomeStatementQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPIncomeStatementQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        base_url = "https://financialmodelingprep.com/stable/income-statement"

        if query.period == "ttm":
            base_url += "-ttm"

        url = (
            base_url
            + f"?symbol={query.symbol}{'&period=' + query.period if query.period != 'ttm' else ''}"
            + f"&limit={query.limit if query.limit else 5}"
            + f"&apikey={api_key}"
        )

        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPIncomeStatementQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPIncomeStatementData]:
        """Return the transformed data."""
        return [FMPIncomeStatementData.model_validate(d) for d in data]



"""FMP Equity Historical Price Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any, Literal

from dateutil.relativedelta import relativedelta
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.equity_historical import (
    EquityHistoricalData,
    EquityHistoricalQueryParams,
)
from openbb_core.provider.utils.descriptions import (
    QUERY_DESCRIPTIONS,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import Field, field_validator, model_validator


class FMPEquityHistoricalQueryParams(EquityHistoricalQueryParams):
    """FMP Equity Historical Price Query.

    Source: https://site.financialmodelingprep.com/developer/docs#historical-price-eod-full
    """

    __alias_dict__ = {"start_date": "from", "end_date": "to"}
    __json_schema_extra__ = {
        "symbol": {"multiple_items_allowed": True},
    }

    interval: Literal["1m", "5m", "15m", "30m", "1h", "4h", "1d"] = Field(
        default="1d", description=QUERY_DESCRIPTIONS.get("interval", "")
    )
    adjustment: Literal["splits_only", "splits_and_dividends", "unadjusted"] = Field(
        default="splits_only",
        description="Type of adjustment for historical prices. Only applies to daily data.",
    )

    @model_validator(mode="before")
    @classmethod
    def _validate_params(cls, values: dict) -> dict:
        """Validate query parameters."""
        interval = values.get("interval", "1d")
        adjustment = values.get("adjustment", "splits_only")

        if adjustment != "splits_only" and interval != "1d":
            raise ValueError("Adjustment can only be applied to daily ('1d') interval.")
        return values


class FMPEquityHistoricalData(EquityHistoricalData):
    """FMP Equity Historical Price Data."""

    __alias_dict__ = {
        "open": "adjOpen",
        "high": "adjHigh",
        "low": "adjLow",
        "close": "adjClose",
    }

    change: float | None = Field(
        default=None,
        description="Change in the price from the previous close.",
    )
    change_percent: float | None = Field(
        default=None,
        description="Change in the price from the previous close, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )

    @field_validator("change_percent", mode="before", check_fields=False)
    @classmethod
    def _normalize_percent(cls, v):
        """Normalize percent."""
        return v / 100 if v else None


class FMPEquityHistoricalFetcher(
    Fetcher[
        FMPEquityHistoricalQueryParams,
        list[FMPEquityHistoricalData],
    ]
):
    """FMP Equity Historical Price Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPEquityHistoricalQueryParams:
        """Transform the query params."""
        transformed_params = params

        now = datetime.now().date()
        if params.get("start_date") is None:
            transformed_params["start_date"] = now - relativedelta(years=1)

        if params.get("end_date") is None:
            transformed_params["end_date"] = now

        return FMPEquityHistoricalQueryParams(**transformed_params)

    @staticmethod
    async def aextract_data(
        query: FMPEquityHistoricalQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_historical_ohlc

        return await get_historical_ohlc(query, credentials, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPEquityHistoricalQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPEquityHistoricalData]:
        """Return the transformed data."""
        if not data:
            raise EmptyDataError("No data returned from FMP for the given query.")
        return [
            FMPEquityHistoricalData.model_validate(d)
            for d in sorted(
                data,
                key=lambda x: (
                    (x["date"], x["symbol"])
                    if len(query.symbol.split(",")) > 1
                    else x["date"]
                ),
                reverse=False,
            )
        ]



"""FMP Equity Ownership Model."""

# pylint: disable=unused-argument

from datetime import date as dateType
from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.equity_ownership import (
    EquityOwnershipData,
    EquityOwnershipQueryParams,
)
from pydantic import Field, field_validator


class FMPEquityOwnershipQueryParams(EquityOwnershipQueryParams):
    """FMP Equity Ownership Query.

    Source: https://site.financialmodelingprep.com/developer/docs#filings-extract-with-analytics-by-holder
    """

    year: int | None = Field(
        default=None,
        description="Calendar year for the data. If not provided, the latest year is used.",
    )
    quarter: int | None = Field(
        default=None,
        description="Calendar quarter for the data. Valid values are 1, 2, 3, or 4."
        + " If not provided, the quarter previous to the current quarter is used.",
        ge=1,
        le=4,
    )
    page: int | None = Field(
        default=None,
        description="Page number, used in conjunction with the limit. The default is 0.",
    )
    limit: int | None = Field(
        default=None,
        description="Number of items to return per page. The default is 100, which is the maximum.",
    )


class FMPEquityOwnershipData(EquityOwnershipData):
    """FMP Equity Ownership Data."""

    __alias_dict__ = {
        "security_type": "typeOfSecurity",
        "sic_industry": "industryTitle",
        "weight_previous": "lastWeight",
        "weight_change": "changeInWeight",
        "weight_change_percent": "changeInWeightPercentage",
        "market_value_previous": "lastMarketValue",
        "market_value_change": "changeInMarketValue",
        "market_value_change_percent": "changeInMarketValuePercentage",
        "shares_previous": "lastSharesNumber",
        "shares_change": "changeInSharesNumber",
        "shares_change_percent": "changeInSharesNumberPercentage",
        "ownership_previous": "lastOwnership",
        "ownership_change": "changeInOwnership",
        "ownership_change_percent": "changeInOwnershipPercentage",
        "performance_percent": "performancePercentage",
        "performance_previous": "lastPerformance",
        "performance_change": "changeInPerformance",
    }

    security_name: str = Field(
        description="Security name.",
    )
    security_type: str = Field(
        description="Type or class of security.",
    )
    security_cusip: str = Field(
        description="CUSIP of the security.",
    )
    shares_type: str = Field(description="Shares type.")
    put_call_share: str = Field(
        description="Whether the share represents a put, call, or share.",
    )
    investment_discretion: str = Field(
        description="Investment discretion of reporting entity.",
    )
    sic_industry: str = Field(
        description="SIC classification industry.",
    )
    weight: float = Field(
        description="Weight relative to the total reported portfolio.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    weight_previous: float = Field(
        description="Weight from the previous quarter.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    weight_change: float = Field(
        description="Change in the weight from the previous quarter.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    weight_change_percent: float = Field(
        description="Change in weight as a percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    market_value: int = Field(description="Market value of the stock ownership.")
    market_value_previous: int = Field(
        description="Market value from the previous quarter.",
    )
    market_value_change: int = Field(
        description="Change in market value from the previous quarter.",
    )
    market_value_change_percent: float = Field(
        description="Change in market value from the previous quarter, as a percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    shares_number: int = Field(
        description="Number of controlled shares.",
    )
    shares_previous: int = Field(
        description="Number of controlled shares from the previous quarter.",
    )
    shares_change: float = Field(
        description="Change in shares number from the previous quarter.",
    )
    shares_change_percent: float = Field(
        description="Change in shares number from the previous quarter, as a percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    quarter_end_price: float = Field(
        description="Market price of the security at the end of the quarter.",
    )
    avg_price_paid: float = Field(
        description="Average price paid for the shares.",
    )
    is_new: bool = Field(description="If the security was newly added this quarter.")
    is_sold_out: bool = Field(description="If the security was sold out this quarter.")
    ownership: float | None = Field(
        default=None,
        description="Ownership stake in the security, as a percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    ownership_previous: float | None = Field(
        default=None,
        description="Ownership stake in the security from the previous quarter, as a percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    ownership_change: float | None = Field(
        default=None,
        description="Change in ownership stake from the previous quarter.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    ownership_change_percent: float | None = Field(
        default=None,
        description="Change in ownership stake from the previous quarter, as a percent.",
    )
    holding_period: int = Field(
        description="Holding period of the security.",
    )
    first_added: dateType = Field(
        description="When the security was first reported as held.",
    )
    performance: float | None = Field(
        default=None, description="Performance value of the security holding."
    )
    performance_percent: float | None = Field(
        default=None,
        description="Performance of the security holding, as a percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    performance_previous: float | None = Field(
        default=None,
        description="Performance value of the security holding from the previous quarter.",
    )
    performance_change: float | None = Field(
        default=None,
        description="Change in value of the security holding's performance.",
    )
    is_counted_for_performance: bool = Field(
        description="If the security is counted for the performance measurement.",
    )

    @field_validator(
        "weight",
        "weight_previous",
        "weight_change",
        "weight_change_percent",
        "market_value_change_percent",
        "shares_change_percent",
        "ownership",
        "ownership_previous",
        "ownership_change",
        "ownership_change_percent",
        "performance_percent",
        mode="before",
        check_fields=False,
    )
    @classmethod
    def _normalize_percent(cls, v):
        """Normalize percent values."""
        return v / 100 if v else None


class FMPEquityOwnershipFetcher(
    Fetcher[
        FMPEquityOwnershipQueryParams,
        list[FMPEquityOwnershipData],
    ]
):
    """FMP Equity Ownership Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPEquityOwnershipQueryParams:
        """Transform the query params."""
        return FMPEquityOwnershipQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPEquityOwnershipQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many
        from pandas import Timestamp, offsets

        api_key = credentials.get("fmp_api_key") if credentials else ""

        year = query.year if query.year else None
        quarter = query.quarter if query.quarter else None

        if year is None and quarter is None:
            current = (Timestamp("now") + offsets.QuarterEnd()) - offsets.QuarterEnd()
            quarter = current.quarter
            year = current.year
        elif year is None and quarter is not None:
            year = Timestamp("now").year
        elif year is not None and quarter is None:
            current = Timestamp("now")
            quarter = (
                4
                if year < current.year
                else current.quarter - 1 if current.quarter > 1 else 1
            )

        url = (
            "https://financialmodelingprep.com/stable/institutional-ownership/extract-analytics/holder"
            + f"?symbol={query.symbol}&year={year}&quarter={quarter}&page={query.page or 0}"
            + f"&limit={query.limit or 100}&apikey={api_key}"
        )

        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPEquityOwnershipQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPEquityOwnershipData]:
        """Return the transformed data."""
        return [FMPEquityOwnershipData.model_validate(d) for d in data]



"""FMP Company Events Calendar Model."""

# pylint: disable=unused-argument

from datetime import date as dateType
from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.calendar_events import (
    CalendarEventsData,
    CalendarEventsQueryParams,
)
from pydantic import Field


class FmpCalendarEventsQueryParams(CalendarEventsQueryParams):
    """FMP Company Events Calendar Query.

    Source: https://site.financialmodelingprep.com/developer/docs/earnings-calendar-confirmed-api
    """

    __alias_dict__ = {"start_date": "from", "end_date": "to"}


class FmpCalendarEventsData(CalendarEventsData):
    """FMP Company Events Calendar Data."""

    __alias_dict__ = {
        "announcement_date": "publicationDate",
        "timing": "when",
        "description": "title",
    }

    exchange: str | None = Field(
        default=None,
        description="Exchange where the symbol is listed.",
    )
    time: str | None = Field(
        default=None,
        description="The estimated time of the event, local to the exchange.",
    )
    timing: str | None = Field(
        default=None,
        description="The timing of the event - e.g. before, during, or after market hours.",
    )
    description: str | None = Field(
        default=None,
        description="The title of the event.",
    )
    url: str | None = Field(
        default=None,
        description="The URL to the press release for the announcement.",
    )
    announcement_date: dateType | None = Field(
        default=None,
        description="The date when the event was announced.",
    )


class FMPCalendarEventsFetcher(
    Fetcher[FmpCalendarEventsQueryParams, list[FmpCalendarEventsData]]
):
    """FMP Company Events Calendar Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FmpCalendarEventsQueryParams:
        """Transform query parameters."""
        return FmpCalendarEventsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FmpCalendarEventsQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list:
        """Extract data from the API."""
        # pylint: disable=import-outside-toplevel
        import asyncio  # noqa
        import warnings
        from datetime import timedelta
        from openbb_core.provider.utils.errors import EmptyDataError, OpenBBError
        from openbb_fmp.utils.helpers import get_data
        from pandas import date_range, to_datetime

        api_key = credentials.get("fmp_api_key") if credentials else ""

        warnings.warn(
            message="This endpoint appears to be deprecated. Please use the earnings calendar endpoint instead.",
        )
        base_url = (
            "https://financialmodelingprep.com/api/v4/earning-calendar-confirmed?"
        )

        start_date = to_datetime(
            query.start_date if query.start_date else dateType.today()
        )
        end_date = to_datetime(
            query.end_date if query.end_date else (dateType.today() + timedelta(days=3))
        )

        # Assuming limit of 1000 events per request, and peak earnings season
        # with 200+ events per day, in America alone, we split into 3-day ranges.
        # We don't actually know what the API limitations are, so this is a conservative guess.
        date_ranges = date_range(start=start_date, end=end_date, freq="3D")
        if end_date not in date_ranges:
            date_ranges = date_ranges.append(to_datetime([end_date]))

        urls: list = []
        results: list = []

        for i in range(len(date_ranges) - 1):
            from_date = date_ranges[i].strftime("%Y-%m-%d")
            to_date = date_ranges[i + 1].strftime("%Y-%m-%d")
            urls.append(
                f"{base_url}from={from_date}&to={to_date}&limit=1000&apikey={api_key}"
            )

        async def get_one(url):
            """Get data from one URL."""
            try:
                response = await get_data(url, **kwargs)
            except OpenBBError as e:
                raise e from e

            if response:
                results.extend(response)

        await asyncio.gather(*[get_one(url) for url in urls])

        if not results:
            raise EmptyDataError("The request was returned empty.")

        return sorted(results, key=lambda x: x["date"])

    @staticmethod
    def transform_data(
        query: FmpCalendarEventsQueryParams, data: list, **kwargs: Any
    ) -> list[FmpCalendarEventsData]:
        """Transform the data."""
        return [FmpCalendarEventsData.model_validate(d) for d in data]



"""FMP Earnings Calendar Model."""

# pylint: disable=unused-argument

from datetime import (
    date as dateType,
    datetime,
    timedelta,
)
from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.calendar_earnings import (
    CalendarEarningsData,
    CalendarEarningsQueryParams,
)
from pydantic import Field, field_validator


class FMPCalendarEarningsQueryParams(CalendarEarningsQueryParams):
    """FMP Earnings Calendar Query.

    Source: https://site.financialmodelingprep.com/developer/docs#earnings-calendar
    """

    __alias_dict__ = {
        "start_date": "from",
        "end_date": "to",
    }


class FMPCalendarEarningsData(CalendarEarningsData):
    """FMP Earnings Calendar Data."""

    __alias_dict__ = {
        "report_date": "date",
        "eps_consensus": "epsEstimated",
        "eps_actual": "epsActual",
        "revenue_actual": "revenueActual",
        "revenue_consensus": "revenueEstimated",
        "last_updated": "lastUpdated",
    }

    eps_actual: float | None = Field(
        default=None,
        description="The actual earnings per share announced.",
    )
    revenue_consensus: float | None = Field(
        default=None,
        description="The revenue forecast consensus.",
    )
    revenue_actual: float | None = Field(
        default=None,
        description="The actual reported revenue.",
    )
    last_updated: dateType | None = Field(
        default=None,
        description="The date the data was updated last.",
    )

    @field_validator(
        "report_date",
        "last_updated",
        mode="before",
        check_fields=False,
    )
    def date_validate(cls, v: datetime | str):  # pylint: disable=E0213
        """Return the date as a datetime object."""
        if isinstance(v, str):
            return datetime.strptime(v, "%Y-%m-%d")
        return datetime.strftime(v, "%Y-%m-%d") if v else None


class FMPCalendarEarningsFetcher(
    Fetcher[
        FMPCalendarEarningsQueryParams,
        list[FMPCalendarEarningsData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPCalendarEarningsQueryParams:
        """Transform the query params."""
        now = datetime.today().date()
        transformed_params = params

        if params.get("start_date") is None:
            transformed_params["start_date"] = now

        if params.get("end_date") is None:
            transformed_params["end_date"] = now + timedelta(days=3)

        return FMPCalendarEarningsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPCalendarEarningsQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_core.provider.utils.helpers import amake_requests
        from openbb_fmp.utils.helpers import response_callback

        api_key = credentials.get("fmp_api_key") if credentials else ""

        base_url = "https://financialmodelingprep.com/stable/earnings-calendar?"
        start_date = query.start_date or datetime.now().date()
        end_date = query.end_date or datetime.now().date() + timedelta(days=7)

        # Create 7-day chunks between start_date and end_date
        urls: list = []
        current_start = start_date

        while current_start <= end_date:
            chunk_end = min(current_start + timedelta(days=7), end_date)
            url = f"{base_url}from={current_start}&to={chunk_end}&apikey={api_key}"
            urls.append(url)
            current_start = chunk_end + timedelta(days=1)

        # Get data from all URLs
        all_data: list = await amake_requests(
            urls, response_callback=response_callback, **kwargs
        )

        return all_data

    @staticmethod
    def transform_data(
        query: FMPCalendarEarningsQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPCalendarEarningsData]:
        """Return the transformed data."""
        data = sorted(data, key=lambda x: x["date"], reverse=True)
        return [FMPCalendarEarningsData.model_validate(d) for d in data]



"""FMP ESG Score Model."""

# pylint: disable=unused-argument

from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.esg_score import (
    EsgScoreData,
    EsgScoreQueryParams,
)


class FMPEsgScoreQueryParams(EsgScoreQueryParams):
    """FMP ESG Score Query."""

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}


class FMPEsgScoreData(EsgScoreData):
    """FMP ESG Score Data."""

    __alias_dict__ = {
        "period_ending": "date",
        "disclosure_date": "acceptedDate",
        "esg_score": "ESGScore",
    }


class FMPEsgScoreFetcher(
    Fetcher[
        FMPEsgScoreQueryParams,
        list[FMPEsgScoreData],
    ]
):
    """FMP ESG Score Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPEsgScoreQueryParams:
        """Transform the query."""
        return FMPEsgScoreQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPEsgScoreQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import asyncio  # noqa
        import warnings
        from openbb_core.provider.utils.errors import EmptyDataError
        from openbb_fmp.utils.helpers import get_data

        api_key = credentials.get("fmp_api_key") if credentials else ""
        symbols = query.symbol.split(",")
        results: list = []

        async def get_one(symbol):
            """Get data for one symbol."""
            url = f"https://financialmodelingprep.com/stable/esg-disclosures?symbol={symbol}&apikey={api_key}"
            result = await get_data(url, **kwargs)

            if not result:
                warnings.warn(f"Symbol Error: No data found for {symbol}")
            elif result:
                results.extend(result)

        await asyncio.gather(*[get_one(symbol) for symbol in symbols])

        if not results:
            raise EmptyDataError("No data found for the given symbols.")

        return sorted(results, key=lambda x: x.get("date", ""), reverse=True)

    @staticmethod
    def transform_data(
        query: FMPEsgScoreQueryParams, data: list, **kwargs: Any
    ) -> list[FMPEsgScoreData]:
        """Return the transformed data."""
        return [FMPEsgScoreData.model_validate(d) for d in data]



"""FMP Historical Employees Model."""

# pylint: disable=unused-argument

from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.historical_employees import (
    HistoricalEmployeesData,
    HistoricalEmployeesQueryParams,
)
from pydantic import ConfigDict, Field


class FMPHistoricalEmployeesQueryParams(HistoricalEmployeesQueryParams):
    """FMP Historical Employees Query.

    Source: https://site.financialmodelingprep.com/developer/docs#historical-employee-count
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}

    limit: int | None = Field(
        default=None,
        description="Number of records to return. Default is all.",
    )


class FMPHistoricalEmployeesData(HistoricalEmployeesData):
    """FMP Historical Employees Data."""

    model_config = ConfigDict(extra="ignore")

    __alias_dict__ = {
        "company_name": "companyName",
        "employees": "employeeCount",
        "date": "periodOfReport",
        "source": "formType",
        "url": "source",
    }

    company_name: str | None = Field(
        default=None,
        description="Company name associated with the data.",
    )
    source: str | None = Field(
        default=None,
        description="Source reference for the data.",
    )
    url: str | None = Field(
        default=None,
        description="URL link to the source of the data.",
    )


class FMPHistoricalEmployeesFetcher(
    Fetcher[
        FMPHistoricalEmployeesQueryParams,
        list[FMPHistoricalEmployeesData],
    ]
):
    """FMP Historical Employees Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPHistoricalEmployeesQueryParams:
        """Transform the query params."""
        return FMPHistoricalEmployeesQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPHistoricalEmployeesQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import warnings  # noqa
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        symbols = query.symbol.split(",")
        results: list = []
        limit = query.limit if query.limit is not None else 10000

        for symbol in symbols:
            url = f"https://financialmodelingprep.com/stable/historical-employee-count?symbol={symbol}&limit={limit}&apikey={api_key}"
            result = await get_data_many(url, **kwargs)

            if not result:
                warnings.warn(f"No data found for symbol {symbol}")
                continue

            results.extend(result)

        return results

    @staticmethod
    def transform_data(
        query: FMPHistoricalEmployeesQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPHistoricalEmployeesData]:
        """Return the transformed data."""
        result: list[FMPHistoricalEmployeesData] = []

        for d in data:
            if query.start_date or query.end_date:
                dt = d.get("periodOfReport")

                if not dt:
                    continue

                if query.start_date and dt < query.start_date.isoformat():
                    continue

                if query.end_date and dt > query.end_date.isoformat():
                    continue

            result.append(FMPHistoricalEmployeesData(**d))

        return sorted(result, key=lambda x: x.date, reverse=True)



"""FMP Price Performance Model."""

# pylint: disable=unused-argument

from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.recent_performance import (
    RecentPerformanceData,
    RecentPerformanceQueryParams,
)
from pydantic import model_validator


class FMPPricePerformanceQueryParams(RecentPerformanceQueryParams):
    """FMP Price Performance Query.

    Source: https://site.financialmodelingprep.com/developer/docs#quote-change
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}


class FMPPricePerformanceData(RecentPerformanceData):
    """FMP Price Performance Data."""

    __alias_dict__ = {
        "one_day": "1D",
        "one_week": "5D",
        "one_month": "1M",
        "three_month": "3M",
        "six_month": "6M",
        "one_year": "1Y",
        "three_year": "3Y",
        "five_year": "5Y",
        "ten_year": "10Y",
    }

    @model_validator(mode="before")
    @classmethod
    def replace_zero(cls, values):  # pylint: disable=no-self-argument
        """Replace zero with None and convert percents to normalized values."""
        if isinstance(values, dict):
            for k, v in values.items():
                if k != "symbol":
                    values[k] = None if v == 0 else float(v) / 100
        return values


class FMPPricePerformanceFetcher(
    Fetcher[
        FMPPricePerformanceQueryParams,
        list[FMPPricePerformanceData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPPricePerformanceQueryParams:
        """Transform the query params."""
        return FMPPricePerformanceQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPPricePerformanceQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_urls

        api_key = credentials.get("fmp_api_key") if credentials else ""
        symbols = query.symbol.upper().split(",")
        chunk_size = 200
        chunks = [
            symbols[i : i + chunk_size] for i in range(0, len(symbols), chunk_size)
        ]
        base_url = "https://financialmodelingprep.com/stable/stock-price-change?symbol="
        urls = [f"{base_url}{','.join(chunk)}&apikey={api_key}" for chunk in chunks]

        return await get_data_urls(urls, **kwargs)  # type: ignore

    @staticmethod
    def transform_data(
        query: FMPPricePerformanceQueryParams,
        data: list,
        **kwargs: Any,
    ) -> list[FMPPricePerformanceData]:
        """Return the transformed data."""
        # pylint: disable=import-outside-toplevel
        import warnings

        symbols = query.symbol.upper().split(",")
        symbols = list(dict.fromkeys(symbols))
        if len(data) != len(symbols):
            data_symbols = [d["symbol"] for d in data]
            missing_symbols = [
                symbol for symbol in symbols if symbol not in data_symbols
            ]
            warnings.warn(f"Missing data for symbols: {missing_symbols}")

        return [FMPPricePerformanceData.model_validate(i) for i in data]



"""FMP Equity Profile Model."""

# pylint: disable=unused-argument

from datetime import (
    date as dateType,
)
from typing import Any

from openbb_core.provider.abstract.data import ForceInt
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.equity_info import (
    EquityInfoData,
    EquityInfoQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import ConfigDict, Field, field_validator, model_validator


class FMPEquityProfileQueryParams(EquityInfoQueryParams):
    """FMP Equity Profile Query.

    Source: https://site.financialmodelingprep.com/developer/docs#profile-symbol
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}


class FMPEquityProfileData(EquityInfoData):
    """FMP Equity Profile Data."""

    __alias_dict__ = {
        "name": "companyName",
        "stock_exchange": "exchange",
        "company_url": "website",
        "hq_address1": "address",
        "hq_address_city": "city",
        "hq_address_postal_code": "zip",
        "hq_state": "state",
        "hq_country": "country",
        "business_phone_no": "phone",
        "industry_category": "industry",
        "employees": "fullTimeEmployees",
        "long_description": "description",
        "first_stock_price_date": "ipoDate",
        "last_price": "price",
        "volume_avg": "averageVolume",
        "annualized_dividend_amount": "lastDividend",
    }
    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}
    model_config = ConfigDict(extra="ignore")

    is_etf: bool = Field(description="If the symbol is an ETF.")
    is_actively_trading: bool = Field(description="If the company is actively trading.")
    is_adr: bool = Field(description="If the stock is an ADR.")
    is_fund: bool = Field(description="If the company is a fund.")
    image: str | None = Field(default=None, description="Image of the company.")
    currency: str | None = Field(
        default=None, description="Currency in which the stock is traded."
    )
    market_cap: ForceInt | None = Field(
        default=None,
        description="Market capitalization of the company.",
    )
    last_price: float | None = Field(
        default=None,
        description="The last traded price.",
    )
    year_high: float | None = Field(
        default=None, description="The one-year high of the price."
    )
    year_low: float | None = Field(
        default=None, description="The one-year low of the price."
    )
    volume_avg: ForceInt | None = Field(
        default=None,
        description="Average daily trading volume.",
    )
    annualized_dividend_amount: float | None = Field(
        default=None,
        description="The annualized dividend payment based on the most recent regular dividend payment.",
    )
    beta: float | None = Field(
        default=None, description="Beta of the stock relative to the market."
    )

    @field_validator("first_stock_price_date", mode="before", check_fields=False)
    @classmethod
    def validate_date(cls, v):  # pylint: disable=E0213
        """Return the date as a datetime object."""
        if isinstance(v, dateType) or v is None:
            return v
        return dateType.fromisoformat(v) if v else None

    @model_validator(mode="before")
    @classmethod
    def replace_empty_strings(cls, values):
        """Check for empty strings and replace with None."""
        return (
            {k: None if v in ("", "NA") else v for k, v in values.items()}
            if isinstance(values, dict)
            else values
        )


class FMPEquityProfileFetcher(
    Fetcher[
        FMPEquityProfileQueryParams,
        list[FMPEquityProfileData],
    ]
):
    """FMP Equity Profile Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPEquityProfileQueryParams:
        """Transform the query params."""
        return FMPEquityProfileQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPEquityProfileQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import asyncio  # noqa
        import warnings
        from openbb_core.provider.utils.helpers import amake_request
        from openbb_fmp.utils.helpers import response_callback

        api_key = credentials.get("fmp_api_key") if credentials else ""
        symbols = query.symbol.split(",")
        base_url = "https://financialmodelingprep.com/stable/"

        results: list = []

        async def get_one(symbol):
            """Get data for one symbol."""
            url = f"{base_url}profile?symbol={symbol}&apikey={api_key}"
            result = await amake_request(
                url, response_callback=response_callback, **kwargs
            )
            if not result:
                warnings.warn(f"Symbol Error: No data found for {symbol}")

            if result:
                results.append(result[0])

        await asyncio.gather(*[get_one(symbol) for symbol in symbols])

        if not results:
            raise EmptyDataError("No data found for the given symbols.")

        return sorted(
            results,
            key=(lambda item: (symbols.index(item.get("symbol", len(symbols))))),
        )

    @staticmethod
    def transform_data(
        query: FMPEquityProfileQueryParams,
        data: list[dict],
        **kwargs: Any,
    ) -> list[FMPEquityProfileData]:
        """Return the transformed data."""
        results: list[FMPEquityProfileData] = []

        for d in data:
            d["year_low"], d["year_high"] = (
                d.pop("range", "-").split("-") if d.get("range") else (None, None)
            )
            results.append(FMPEquityProfileData.model_validate(d))

        return results



"""FMP Forward EPS Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any, Literal

from openbb_core.app.model.abstract.error import OpenBBError
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.forward_eps_estimates import (
    ForwardEpsEstimatesData,
    ForwardEpsEstimatesQueryParams,
)
from openbb_core.provider.utils.descriptions import QUERY_DESCRIPTIONS
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import Field, field_validator


class FMPForwardEpsEstimatesQueryParams(ForwardEpsEstimatesQueryParams):
    """FMP Forward EPS Query.

    Source: https://site.financialmodelingprep.com/developer/docs/analyst-estimates-api/
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}

    __alias_dict__ = {"fiscal_period": "period"}

    fiscal_period: Literal["annual", "quarter"] = Field(
        default="annual",
        description="The future fiscal period to retrieve estimates for.",
    )
    limit: int | None = Field(
        default=None,
        description=QUERY_DESCRIPTIONS.get("limit", "")
        + " Number of historical periods.",
    )
    include_historical: bool = Field(
        default=False,
        description="If True, the data will include all past data and the limit will be ignored.",
    )

    @field_validator("symbol", mode="before", check_fields=False)
    @classmethod
    def check_symbol(cls, value):
        """Check the symbol."""
        if not value:
            raise OpenBBError("Symbol is a required field for FMP.")
        return value


class FMPForwardEpsEstimatesData(ForwardEpsEstimatesData):
    """FMP Forward EPS Data."""

    __alias_dict__ = {
        "number_of_analysts": "numberAnalystsEps",
        "high_estimate": "epsHigh",
        "low_estimate": "epsLow",
        "mean": "epsAvg",
    }


class FMPForwardEpsEstimatesFetcher(
    Fetcher[
        FMPForwardEpsEstimatesQueryParams,
        list[FMPForwardEpsEstimatesData],
    ]
):
    """FMP Forward EPS Estimates Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPForwardEpsEstimatesQueryParams:
        """Transform the query params."""
        return FMPForwardEpsEstimatesQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPForwardEpsEstimatesQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import asyncio  # noqa
        import warnings
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        symbols = query.symbol.split(",")  # type: ignore
        results: list[dict] = []
        base_url = "https://financialmodelingprep.com/stable/analyst-estimates?"
        limit = query.limit if query.limit else 1000

        async def get_one(symbol):
            """Get data for one symbol."""
            url = f"{base_url}symbol={symbol}&period={query.fiscal_period}&limit={limit}&apikey={api_key}"
            result = await get_data_many(url, **kwargs)
            if not result or len(result) == 0:
                warnings.warn(f"Symbol Error: No data found for {symbol}")
            if result:
                results.extend(result)

        await asyncio.gather(*[get_one(symbol) for symbol in symbols])

        if not results:
            raise EmptyDataError("No data returned for the given symbols.")

        return results

    @staticmethod
    def transform_data(
        query: FMPForwardEpsEstimatesQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPForwardEpsEstimatesData]:
        """Return the transformed data."""
        symbols = query.symbol.split(",") if query.symbol else []
        cols = [
            "symbol",
            "date",
            "epsAvg",
            "epsHigh",
            "epsLow",
            "numberAnalystsEps",
        ]
        year = datetime.now().year
        results: list[FMPForwardEpsEstimatesData] = []

        for item in sorted(
            data,
            key=lambda item: (  # type: ignore
                (
                    symbols.index(item.get("symbol")) if item.get("symbol") in symbols else len(symbols),  # type: ignore
                    item.get("date"),
                )
                if symbols
                else item.get("date")
            ),
        ):
            temp: dict[str, Any] = {}

            for col in cols:
                temp[col] = item.get(col)

            if (
                query.include_historical is False
                and datetime.strptime(temp["date"], "%Y-%m-%d").year < year
            ):
                continue

            results.append(FMPForwardEpsEstimatesData.model_validate(temp))

        return results



"""FMP Executive Compensation Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.executive_compensation import (
    ExecutiveCompensationData,
    ExecutiveCompensationQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import Field


class FMPExecutiveCompensationQueryParams(ExecutiveCompensationQueryParams):
    """FMP Executive Compensation Query.

    Source: https://site.financialmodelingprep.com/developer/docs#executive-compensation
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}

    year: int = Field(
        default=-1,
        description="Filters results by year, enter 0 for all data available."
        + " Default is the most recent year in the dataset, -1.",
    )


class FMPExecutiveCompensationData(ExecutiveCompensationData):
    """FMP Executive Compensation Data."""

    __alias_dict__ = {
        "company_name": "companyName",
        "industry": "industryTitle",
        "url": "link",
        "executive": "nameAndPosition",
        "report_date": "filingDate",
    }
    accepted_date: datetime | None = Field(
        default=None, description="Date the filing was accepted."
    )
    url: str | None = Field(default=None, description="URL to the filing data.")


class FMPExecutiveCompensationFetcher(
    Fetcher[
        FMPExecutiveCompensationQueryParams,
        list[FMPExecutiveCompensationData],
    ]
):
    """FMP Executive Compensation Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPExecutiveCompensationQueryParams:
        """Transform the query params."""
        return FMPExecutiveCompensationQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPExecutiveCompensationQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import asyncio  # noqa
        import warnings
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        base_url = "https://financialmodelingprep.com/stable/governance-executive-compensation?"
        symbols = query.symbol.split(",")
        results: list = []

        async def get_one(symbol):
            """Get data for one symbol."""
            url = f"{base_url}symbol={symbol}&apikey={api_key}"
            result = await get_data_many(url, **kwargs)

            if not result:
                warnings.warn(f"Symbol Error: No data found for {symbol}.")

            if result:
                results.extend(result)

        await asyncio.gather(*[get_one(symbol) for symbol in symbols])

        if not results:
            raise EmptyDataError("No data found for given symbols.")

        return results

    @staticmethod
    def transform_data(
        query: FMPExecutiveCompensationQueryParams,
        data: list,
        **kwargs: Any,
    ) -> list[FMPExecutiveCompensationData]:
        """Return the transformed data."""
        # pylint: disable=import-outside-toplevel
        import warnings

        symbols = query.symbol.split(",")
        filtered_results: list[FMPExecutiveCompensationData] = []

        for symbol in symbols:
            symbol_data = [item for item in data if item.get("symbol") == symbol]

            if symbol_data and query.year != 0:
                max_year_for_symbol = (
                    (
                        max(
                            item.get("year", 0)
                            for item in symbol_data
                            if item.get("year")
                        )
                    )
                    if query.year == -1
                    else query.year
                )
                symbol_max_year_data = [
                    item
                    for item in symbol_data
                    if int(item.get("year", 0)) == max_year_for_symbol
                ]
                if not symbol_max_year_data:
                    warnings.warn(
                        f"ValueError: No data found for {symbol} and year {query.year}."
                    )
                    continue

                filtered_results.extend(
                    [
                        FMPExecutiveCompensationData.model_validate(item)
                        for item in symbol_max_year_data
                    ]
                )
            else:
                filtered_results.extend(
                    [
                        FMPExecutiveCompensationData.model_validate(item)
                        for item in sorted(
                            symbol_data, key=lambda x: x.get("year", 0), reverse=True
                        )
                    ]
                )

        if not filtered_results:
            raise EmptyDataError("No data found for given symbols and year.")

        return filtered_results



"""FMP Income Statement Growth Model."""

# pylint: disable=unused-argument

from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.income_statement_growth import (
    IncomeStatementGrowthData,
    IncomeStatementGrowthQueryParams,
)
from openbb_core.provider.utils.descriptions import (
    DATA_DESCRIPTIONS,
    QUERY_DESCRIPTIONS,
)
from openbb_fmp.utils.definitions import FinancialPeriods
from pydantic import Field


class FMPIncomeStatementGrowthQueryParams(IncomeStatementGrowthQueryParams):
    """FMP Income Statement Growth Query.

    Source: https://site.financialmodelingprep.com/developer/docs#income-statement-growth
    """

    period: FinancialPeriods = Field(
        default="annual",
        description=QUERY_DESCRIPTIONS.get("period", ""),
    )


class FMPIncomeStatementGrowthData(IncomeStatementGrowthData):
    """FMP Income Statement Growth Data."""

    __alias_dict__ = {
        "period_ending": "date",
        "fiscal_year": "calendarYear",
        "fiscal_period": "period",
        "growth_ebit": "growthEBIT",
        "growth_ebitda": "growthEBITDA",
        "growth_basic_earings_per_share": "growthEPS",
        "growth_gross_profit_margin": "growthGrossProfitRatio",
        "growth_consolidated_net_income": "growthNetIncome",
        "growth_diluted_earnings_per_share": "growthEPSDiluted",
        "growth_weighted_average_basic_shares_outstanding": "growthWeightedAverageShsOut",
        "growth_weighted_average_diluted_shares_outstanding": "growthWeightedAverageShsOutDil",
        "growth_research_and_development_expense": "growthResearchAndDevelopmentExpenses",
        "growth_general_and_admin_expense": "growthGeneralAndAdministrativeExpenses",
        "growth_selling_and_marketing_expense": "growthSellingAndMarketingExpenses",
    }

    symbol: str = Field(description=DATA_DESCRIPTIONS.get("symbol", ""))
    reported_currency: str | None = Field(
        description="The currency in which the financial data is reported.",
        default=None,
    )
    growth_revenue: float | None = Field(
        default=None,
        description="Growth rate of total revenue.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_cost_of_revenue: float | None = Field(
        default=None,
        description="Growth rate of cost of goods sold.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_gross_profit: float | None = Field(
        default=None,
        description="Growth rate of gross profit.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_gross_profit_margin: float | None = Field(
        default=None,
        description="Growth rate of gross profit as a percentage of revenue.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_general_and_admin_expense: float | None = Field(
        default=None,
        description="Growth rate of general and administrative expenses.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_research_and_development_expense: float | None = Field(
        default=None,
        description="Growth rate of expenses on research and development.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_selling_and_marketing_expense: float | None = Field(
        default=None,
        description="Growth rate of expenses on selling and marketing activities.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_other_expenses: float | None = Field(
        default=None,
        description="Growth rate of other operating expenses.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_operating_expenses: float | None = Field(
        default=None,
        description="Growth rate of total operating expenses.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_cost_and_expenses: float | None = Field(
        default=None,
        description="Growth rate of total costs and expenses.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_depreciation_and_amortization: float | None = Field(
        default=None,
        description="Growth rate of depreciation and amortization expenses.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_interest_income: float | None = Field(
        default=None,
        description="Growth rate of interest income.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_interest_expense: float | None = Field(
        default=None,
        description="Growth rate of interest expenses.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_net_interest_income: float | None = Field(
        default=None,
        description="Growth rate of net interest income.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_ebit: float | None = Field(
        default=None,
        description="Growth rate of Earnings Before Interest and Taxes (EBIT).",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_ebitda: float | None = Field(
        default=None,
        description="Growth rate of Earnings Before Interest, Taxes, Depreciation, and Amortization.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_operating_income: float | None = Field(
        default=None,
        description="Growth rate of operating income.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_non_operating_income_excluding_interest: float | None = Field(
        default=None,
        description="Growth rate of non-operating income excluding interest.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_total_other_income_expenses_net: float | None = Field(
        default=None,
        description="Growth rate of net total other income and expenses.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_other_adjustments_to_net_income: float | None = Field(
        default=None,
        description="Growth rate of other adjustments to net income.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_net_income_deductions: float | None = Field(
        default=None,
        description="Growth rate of net income deductions.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_income_before_tax: float | None = Field(
        default=None,
        description="Growth rate of income before taxes.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_income_tax_expense: float | None = Field(
        default=None,
        description="Growth rate of income tax expenses.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_net_income_from_continuing_operations: float | None = Field(
        default=None,
        description="Growth rate of net income from continuing operations.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_consolidated_net_income: float | None = Field(
        default=None,
        description="Growth rate of net income.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_basic_earings_per_share: float | None = Field(
        default=None,
        description="Growth rate of Earnings Per Share (EPS).",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_diluted_earnings_per_share: float | None = Field(
        default=None,
        description="Growth rate of diluted Earnings Per Share (EPS).",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_weighted_average_basic_shares_outstanding: float | None = Field(
        default=None,
        description="Growth rate of weighted average shares outstanding.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_weighted_average_diluted_shares_outstanding: float | None = Field(
        default=None,
        description="Growth rate of diluted weighted average shares outstanding.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )


class FMPIncomeStatementGrowthFetcher(
    Fetcher[
        FMPIncomeStatementGrowthQueryParams,
        list[FMPIncomeStatementGrowthData],
    ]
):
    """FMP Income Statement Growth Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPIncomeStatementGrowthQueryParams:
        """Transform the query params."""
        return FMPIncomeStatementGrowthQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPIncomeStatementGrowthQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""

        url = (
            "https://financialmodelingprep.com/stable/income-statement-growth"
            + f"?symbol={query.symbol}"
            + f"&period={query.period}"
            + f"&limit={query.limit if query.limit else 5}"
            + f"&apikey={api_key}"
        )

        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPIncomeStatementGrowthQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPIncomeStatementGrowthData]:
        """Return the transformed data."""
        return [FMPIncomeStatementGrowthData.model_validate(d) for d in data]



"""FMP Currency Snapshots Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.currency_snapshots import (
    CurrencySnapshotsData,
    CurrencySnapshotsQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import Field, field_validator


class FMPCurrencySnapshotsQueryParams(CurrencySnapshotsQueryParams):
    """FMP Currency Snapshots Query.

    Source: https://site.financialmodelingprep.com/developer/docs#all-forex-quotes
    """

    __json_schema_extra__ = {"base": {"multiple_items_allowed": True}}


class FMPCurrencySnapshotsData(CurrencySnapshotsData):
    """FMP Currency Snapshots Data."""

    __alias_dict__ = {
        "last_rate": "price",
        "high": "dayHigh",
        "low": "dayLow",
        "ma50": "priceAvg50",
        "ma200": "priceAvg200",
        "year_high": "yearHigh",
        "year_low": "yearLow",
        "prev_close": "previousClose",
        "change_percent": "changePercentage",
        "last_rate_timestamp": "timestamp",
    }

    change: float | None = Field(
        description="The change in the price from the previous close.", default=None
    )
    change_percent: float | None = Field(
        description="The change in the price from the previous close, as a normalized percent.",
        default=None,
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    ma50: float | None = Field(description="The 50-day moving average.", default=None)
    ma200: float | None = Field(description="The 200-day moving average.", default=None)
    year_high: float | None = Field(description="The 52-week high.", default=None)
    year_low: float | None = Field(description="The 52-week low.", default=None)
    last_rate_timestamp: datetime | None = Field(
        description="The timestamp of the last rate.", default=None
    )

    @field_validator("change_percent", mode="before", check_fields=False)
    @classmethod
    def normalize_percent(cls, v):
        """Normalize the percent."""
        return v / 100 if v is not None else None


class FMPCurrencySnapshotsFetcher(
    Fetcher[FMPCurrencySnapshotsQueryParams, list[FMPCurrencySnapshotsData]]
):
    """FMP Currency Snapshots Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPCurrencySnapshotsQueryParams:
        """Transform the query parameters."""
        return FMPCurrencySnapshotsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPCurrencySnapshotsQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list[dict]:
        """Extract the data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""

        url = f"https://financialmodelingprep.com/stable/batch-forex-quotes?short=false&apikey={api_key}"

        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPCurrencySnapshotsQueryParams,
        data: list[dict],
        **kwargs: Any,
    ) -> list[FMPCurrencySnapshotsData]:
        """Filter by the query parameters and validate the model."""
        # pylint: disable=import-outside-toplevel
        from datetime import timezone  # noqa
        from numpy import nan
        from pandas import DataFrame, concat
        from openbb_core.provider.utils.helpers import safe_fromtimestamp

        if not data:
            raise EmptyDataError("No data was returned from the FMP endpoint.")

        # Drop all the zombie columns FMP returns.
        df = DataFrame(data).dropna(how="all", axis=1).drop(columns=["exchange"])

        new_df = DataFrame()

        # Filter for the base currencies requested and the quote_type.
        for symbol in query.base.split(","):
            temp = (
                df.query("`symbol`.str.startswith(@symbol)")
                if query.quote_type == "indirect"
                else df.query("`symbol`.str.endswith(@symbol)")
            ).rename(columns={"symbol": "base_currency", "name": "counter_currency"})
            temp["base_currency"] = symbol
            temp["counter_currency"] = (
                [d.split("/")[1] for d in temp["counter_currency"]]
                if query.quote_type == "indirect"
                else [d.split("/")[0] for d in temp["counter_currency"]]
            )
            # Filter for the counter currencies, if requested.
            if query.counter_currencies is not None:
                counter_currencies = (  # noqa: F841  # pylint: disable=unused-variable
                    query.counter_currencies
                    if isinstance(query.counter_currencies, list)
                    else query.counter_currencies.split(",")
                )
                temp = (
                    temp.query("`counter_currency`.isin(@counter_currencies)")
                    .set_index("counter_currency")
                    # Sets the counter currencies in the order they were requested.
                    .filter(items=counter_currencies, axis=0)
                    .reset_index()
                ).rename(columns={"index": "counter_currency"})
            # If there are no records, don't concatenate.
            if len(temp) > 0:
                # Convert the Unix timestamp to a datetime.
                temp.timestamp = temp.timestamp.apply(
                    lambda x: safe_fromtimestamp(x, tz=timezone.utc)
                )
                new_df = concat([new_df, temp])
            if len(new_df) == 0:
                raise EmptyDataError(
                    "No data was found using the applied filters. Check the parameters."
                )
            new_df = new_df.replace({nan: None})

        return [
            FMPCurrencySnapshotsData.model_validate(d)
            for d in new_df.reset_index(drop=True).to_dict(orient="records")
        ]



"""FMP Index Historical Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any, Literal

from dateutil.relativedelta import relativedelta
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.index_historical import (
    IndexHistoricalData,
    IndexHistoricalQueryParams,
)
from openbb_core.provider.utils.descriptions import (
    DATA_DESCRIPTIONS,
    QUERY_DESCRIPTIONS,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import Field, field_validator


class FMPIndexHistoricalQueryParams(IndexHistoricalQueryParams):
    """FMP Index Historical Query.

    Source: https://site.financialmodelingprep.com/developer/docs/historical-index-price-api/
    """

    __alias_dict__ = {"start_date": "from", "end_date": "to"}
    __json_schema_extra__ = {
        "symbol": {"multiple_items_allowed": True},
    }

    interval: Literal["1m", "5m", "1h", "1d"] = Field(
        default="1d", description=QUERY_DESCRIPTIONS.get("interval", "")
    )


class FMPIndexHistoricalData(IndexHistoricalData):
    """FMP Index Historical Data."""

    vwap: float | None = Field(
        default=None, description=DATA_DESCRIPTIONS.get("vwap", "")
    )
    change: float | None = Field(
        default=None,
        description="Change in the price from the previous close.",
    )
    change_percent: float | None = Field(
        default=None,
        description="Change in the price from the previous close, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )

    @field_validator("change_percent", mode="before", check_fields=False)
    @classmethod
    def _normalize_percent(cls, v):
        """Normalize percent values."""
        return v / 100 if v else None


class FMPIndexHistoricalFetcher(
    Fetcher[
        FMPIndexHistoricalQueryParams,
        list[FMPIndexHistoricalData],
    ]
):
    """FMP Index Historical Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPIndexHistoricalQueryParams:
        """Transform the query params."""
        transformed_params = params

        now = datetime.now().date()
        if params.get("start_date") is None:
            transformed_params["start_date"] = now - relativedelta(years=1)

        if params.get("end_date") is None:
            transformed_params["end_date"] = now

        return FMPIndexHistoricalQueryParams.model_validate(transformed_params)

    @staticmethod
    async def aextract_data(
        query: FMPIndexHistoricalQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_historical_ohlc

        return await get_historical_ohlc(query, credentials, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPIndexHistoricalQueryParams,
        data: list[dict],
        **kwargs: Any,
    ) -> list[FMPIndexHistoricalData]:
        """Return the transformed data."""
        if not data:
            raise EmptyDataError()

        return [
            FMPIndexHistoricalData.model_validate(d)
            for d in sorted(
                data,
                key=lambda x: (
                    (x["date"], x["symbol"])
                    if len(query.symbol.split(",")) > 1
                    else x["date"]
                ),
                reverse=False,
            )
        ]



"""FMP Earnings Call Transcript Model."""

# pylint: disable=unused-argument
from typing import Any

from openbb_core.app.model.abstract.error import OpenBBError
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.earnings_call_transcript import (
    EarningsCallTranscriptData,
    EarningsCallTranscriptQueryParams,
)
from pydantic import ConfigDict


class FMPEarningsCallTranscriptQueryParams(EarningsCallTranscriptQueryParams):
    """FMP Earnings Call Transcript Query.

    Source: https://site.financialmodelingprep.com/developer/docs/earning-call-transcript-api/
    """


class FMPEarningsCallTranscriptData(EarningsCallTranscriptData):
    """FMP Earnings Call Transcript Data."""

    model_config = ConfigDict(
        json_schema_extra={
            "x-widget_config": {
                "$.name": "Earnings Call Transcript",
                "$.description": "Earnings calls by symbol, year, and quarter.",
                "$.type": "markdown",
                "$.gridData": {
                    "h": 20,
                    "w": 40,
                },
                "$.data": {
                    "dataKey": "results.content",
                },
                "$.params": [
                    {
                        "paramName": "year",
                        "label": "Fiscal Year",
                        "type": "number",
                        "description": "Fiscal year of the earnings call transcript."
                        + " When not provided, or not available, the most recent will be used.",
                    },
                    {
                        "paramName": "quarter",
                        "label": "Fiscal Quarter",
                        "description": "Fiscal quarter of the earnings call transcript."
                        + " When not provided, the latest quarter available will be used.",
                    },
                ],
                "$.refetchInterval": False,
            }
        }
    )

    __alias_dict__ = {
        "quarter": "period",
    }


class FMPEarningsCallTranscriptFetcher(
    Fetcher[
        FMPEarningsCallTranscriptQueryParams,
        FMPEarningsCallTranscriptData,
    ]
):
    """FMP Earnings Call Transcript Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPEarningsCallTranscriptQueryParams:
        """Transform the query params."""
        return FMPEarningsCallTranscriptQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPEarningsCallTranscriptQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> dict:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import warnings  # noqa
        from openbb_fmp.utils.helpers import (
            get_available_transcript_symbols,
            get_data_one,
            get_transcript_dates_for_symbol,
        )
        from pandas import DataFrame

        api_key = credentials.get("fmp_api_key") if credentials else ""

        available_symbols = get_available_transcript_symbols(api_key=api_key)
        avail_df = DataFrame(available_symbols)

        if query.symbol.upper() not in avail_df["symbol"].values:
            raise OpenBBError(
                ValueError(
                    f"Symbol {query.symbol} not found in available transcripts."
                    + f"\n Available symbols include: {', '.join(sorted(avail_df['symbol'].unique().tolist()))}"
                )
            )
        symbol_transcripts = get_transcript_dates_for_symbol(
            query.symbol.upper(), api_key=api_key
        )

        df_dates = DataFrame(symbol_transcripts).sort_values(by="date", ascending=False)
        year = df_dates.iloc[0].fiscalYear

        if query.year and query.year not in df_dates.fiscalYear.values:
            warnings.warn(
                f"Year {query.year} not found in available transcripts for {query.symbol}."
                + f" Using latest year {year} instead."
            )

        year = query.year if query.year in df_dates.fiscalYear.values else year

        quarter = query.quarter if query.quarter else df_dates.iloc[0].quarter

        if (
            query.quarter
            and query.quarter
            not in df_dates.query("fiscalYear == @year").quarter.values
        ):
            warnings.warn(
                f"Quarter {query.quarter} not found in available transcripts for {query.symbol} in {year}."
                + f" Using latest quarter q{df_dates.query('fiscalYear == @year').iloc[0].quarter} instead."
            )

        url = (
            "https://financialmodelingprep.com/stable/earning-call-transcript?symbol="
            + f"{query.symbol.upper()}&year={year}&quarter={quarter}&apikey={api_key}"
        )

        try:
            return await get_data_one(url, **kwargs)
        except ValueError as e:
            raise OpenBBError(
                f"No transcript found for {query.symbol} in {year} Q{quarter}"
                f". \n Latest available transcript is {df_dates.iloc[0].fiscalYear} Q{df_dates.iloc[0].quarter}."
            ) from e

    @staticmethod
    def transform_data(
        query: FMPEarningsCallTranscriptQueryParams, data: dict, **kwargs: Any
    ) -> FMPEarningsCallTranscriptData:
        """Return the transformed data."""
        if not data:
            raise OpenBBError(
                ValueError(
                    f"No data found for {query.symbol} for year {query.year} and period {query.quarter}."
                )
            )
        transcript = data.get("content", "")

        output_lines: list = []
        intro_lines = f"""
## {data.get("symbol")} - {data.get("year")} {data.get("period")} Earnings Call Transcript - {data.get("date")}
\n\n
"""
        output_lines.append(intro_lines + "\n\n")
        for line in transcript.splitlines():
            section_title = line.split(":", 1)[0] if ":" in line else ""
            section_line = line.split(":", 1)[1] if ":" in line else ""
            if section_title and section_line:
                output_lines.append(f"### **{section_title.strip()}**:" + "\n\n")
                output_lines.append(section_line.strip() + "\n\n")
            else:
                output_lines.append(line)

        data["content"] = "\n".join(output_lines)

        return FMPEarningsCallTranscriptData.model_validate(data)



"""FMP ETF Search Model."""

# pylint: disable=unused-argument

from typing import Any, Literal

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.etf_search import (
    EtfSearchData,
    EtfSearchQueryParams,
)
from openbb_core.provider.utils.country_utils import Country
from openbb_fmp.utils.definitions import Countries
from pydantic import ConfigDict, Field, field_validator


class FMPEtfSearchQueryParams(EtfSearchQueryParams):
    """FMP ETF Search Query."""

    __json_schema_extra__ = {
        "exchange": {
            "x-widget_config": {
                "options": [
                    {"label": "AMEX", "value": "amex"},
                    {"label": "NYSE", "value": "nyse"},
                    {"label": "NASDAQ", "value": "nasdaq"},
                    {"label": "TSX", "value": "tsx"},
                    {"label": "Euronext", "value": "euronext"},
                ],
            }
        },
        "country": {
            "x-widget_config": {
                "options": [
                    {"label": country.upper(), "value": country}
                    for country in Countries.__args__
                ],
            },
        },
    }

    exchange: Literal["amex", "nyse", "nasdaq", "tsx", "euronext"] | None = Field(
        description="Exchange where the ETF is listed. If not provided, all exchanges are searched.",
        default=None,
    )
    country: Country | None = Field(
        default=None,
        description="Filter by country. Accepts ISO 3166-1 alpha-2 codes (e.g., 'US', 'DE'), "
        "alpha-3 codes (e.g., 'USA'), or country names (e.g., 'United States', 'united_states').",
    )

    @field_validator("country", mode="after")
    @classmethod
    def _validate_country(cls, v):
        """Validate country is supported by FMP."""
        if v is None:
            return v
        # Country stores alpha_2 in uppercase, FMP expects lowercase
        country_code = v.alpha_2.lower()
        valid_countries = list(Countries.__args__)
        if country_code not in valid_countries:
            raise ValueError(
                f"Country '{v.name}' ({v.alpha_2}) is not supported by FMP. "
                f"Valid options: {', '.join(sorted(valid_countries)[:20])}..."
            )
        return v


class FMPEtfSearchData(EtfSearchData):
    """FMP ETF Search Data."""

    model_config = ConfigDict(extra="ignore")

    __alias_dict__ = {
        "name": "companyName",
        "market_cap": "marketCap",
        "last_annual_dividend": "lastAnnualDividend",
        "exchange": "exchangeShortName",
        "exchange_name": "exchange",
    }
    country: str | None = Field(
        description="Country where the ETF is domiciled.", default=None
    )
    exchange: str | None = Field(
        description="Exchange where the ETF is listed.",
        default=None,
    )
    exchange_name: str | None = Field(
        description="The full name of the exchange.",
        default=None,
    )
    market_cap: int | float | None = Field(
        description="Market capitalization of the ETF.", default=None
    )
    beta: float | None = Field(description="Beta of the ETF.", default=None)
    price: float | None = Field(description="Current price of the ETF.", default=None)
    last_annual_dividend: float | None = Field(
        description="Last annual dividend paid.",
        default=None,
    )
    volume: int | float | None = Field(
        description="Current trading volume of the ETF.", default=None
    )


class FMPEtfSearchFetcher(
    Fetcher[
        FMPEtfSearchQueryParams,
        list[FMPEtfSearchData],
    ]
):
    """FMP ETF Search Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPEtfSearchQueryParams:
        """Transform the query."""
        return FMPEtfSearchQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPEtfSearchQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""

        url = "https://financialmodelingprep.com/stable/company-screener?isEtf=true&isFund=false&isActivelyTrading=true"
        if query.exchange:
            url += f"&exchange={query.exchange.upper()}"
        if query.country:
            url += f"&country={query.country.upper()}"

        url += f"&limit=10000&apikey={api_key}"

        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPEtfSearchQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPEtfSearchData]:
        """Return the transformed data."""
        # pylint: disable=import-outside-toplevel
        from numpy import nan
        from pandas import DataFrame

        etfs = DataFrame(data)

        if query.query:
            etfs = etfs[
                etfs["companyName"].str.contains(query.query, case=False)
                | etfs["exchangeShortName"].str.contains(query.query, case=False)
                | etfs["exchange"].str.contains(query.query, case=False)
                | etfs["symbol"].str.contains(query.query, case=False)
            ]

        etfs = etfs.replace(
            {
                nan: None,
                "": None,
                0: None,
            }
        )
        return [FMPEtfSearchData.model_validate(d) for d in etfs.to_dict("records")]



"""FMP Treasury Rates Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any

from dateutil.relativedelta import relativedelta
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.treasury_rates import (
    TreasuryRatesData,
    TreasuryRatesQueryParams,
)
from pydantic import model_validator


class FMPTreasuryRatesQueryParams(TreasuryRatesQueryParams):
    """FMP Treasury Rates Query.

    Source: https://site.financialmodelingprep.com/developer/docs#treasury-rates
    """


class FMPTreasuryRatesData(TreasuryRatesData):
    """FMP Treasury Rates Data."""

    __alias_dict__ = {
        "month_1": "month1",
        "month_2": "month2",
        "month_3": "month3",
        "month_6": "month6",
        "year_1": "year1",
        "year_2": "year2",
        "year_3": "year3",
        "year_5": "year5",
        "year_7": "year7",
        "year_10": "year10",
        "year_20": "year20",
        "year_30": "year30",
    }

    @model_validator(mode="before")
    @classmethod
    def normalize_percent(cls, values):
        """Normalize the percent values."""
        for k, v in values.items():
            if k != "date" and v:
                values[k] = float(v) / 100
        return values


class FMPTreasuryRatesFetcher(
    Fetcher[
        FMPTreasuryRatesQueryParams,
        list[FMPTreasuryRatesData],
    ]
):
    """FMP Treasury Rates Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPTreasuryRatesQueryParams:
        """Transform the query params. Start and end dates are set to a 90 day interval."""
        transformed_params = params

        now = datetime.now().date()
        if params.get("start_date") is None:
            transformed_params["start_date"] = now - relativedelta(years=1)

        if params.get("end_date") is None:
            transformed_params["end_date"] = now

        return FMPTreasuryRatesQueryParams(**transformed_params)

    @staticmethod
    async def aextract_data(
        query: FMPTreasuryRatesQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_urls

        api_key = credentials.get("fmp_api_key") if credentials else ""

        def generate_urls(start_date, end_date):
            """Generate URLs for each 3-month interval between start_date and end_date."""
            base_url = (
                "https://financialmodelingprep.com/stable/treasury-rates?from={}&to={}"
            )
            urls = []
            while start_date <= end_date:
                next_date = start_date + relativedelta(months=3)
                url = base_url.format(
                    start_date.strftime("%Y-%m-%d"),
                    min(next_date, end_date).strftime("%Y-%m-%d"),
                )
                url = url + f"&apikey={api_key}"
                urls.append(url)
                start_date = next_date
            return urls

        urls = generate_urls(query.start_date, query.end_date)

        return await get_data_urls(urls, **kwargs)  # type: ignore

    @staticmethod
    def transform_data(
        query: FMPTreasuryRatesQueryParams, data: list, **kwargs: Any
    ) -> list[FMPTreasuryRatesData]:
        """Return the transformed data."""
        return [
            FMPTreasuryRatesData.model_validate(d)
            for d in sorted(data, key=lambda x: x["date"], reverse=False)
        ]



"""FMP Crypto Search Model."""

# pylint: disable=unused-argument

from datetime import date as dateType
from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.crypto_search import (
    CryptoSearchData,
    CryptoSearchQueryParams,
)
from pydantic import Field, field_validator


class FMPCryptoSearchQueryParams(CryptoSearchQueryParams):
    """FMP Crypto Search Query.

    Source: https://site.financialmodelingprep.com/developer/docs#cryptocurrency-list
    """

    @field_validator("query", mode="after", check_fields=False)
    def validate_query(cls, v: str) -> str:  # pylint: disable=no-self-argument
        """Return the query."""
        if isinstance(v, str):
            return v.replace("-", "") if "-" in v else v
        return None


class FMPCryptoSearchData(CryptoSearchData):
    """FMP Crypto Search Data."""

    exchange: str | None = Field(
        description="The exchange code the crypto trades on.",
        default=None,
    )
    ico_date: dateType | None = Field(
        description="The ICO date of the token.",
        default=None,
    )
    circulating_supply: float | None = Field(
        description="The circulating supply of the token.",
        default=None,
    )
    total_supply: float | None = Field(
        description="The total supply of the token.",
        default=None,
    )


class FMPCryptoSearchFetcher(
    Fetcher[
        FMPCryptoSearchQueryParams,
        list[FMPCryptoSearchData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPCryptoSearchQueryParams:
        """Transform the query."""
        return FMPCryptoSearchQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPCryptoSearchQueryParams,  # pylint: disable=unused-argument
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""

        url = f"https://financialmodelingprep.com/stable/cryptocurrency-list?apikey={api_key}"

        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPCryptoSearchQueryParams,
        data: list[dict],
        **kwargs: Any,
    ) -> list[FMPCryptoSearchData]:
        """Return the transformed data."""
        # pylint: disable=import-outside-toplevel
        from pandas import DataFrame

        cryptos = DataFrame(data)
        if query.query:
            cryptos = cryptos[
                cryptos["symbol"].str.contains(query.query, case=False)
                | cryptos["name"].str.contains(query.query, case=False)
                | cryptos["exchange"].str.contains(query.query, case=False)
            ]
        return [
            FMPCryptoSearchData.model_validate(d) for d in cryptos.to_dict("records")
        ]



"""FMP Historical EPS Model."""

# pylint: disable=unused-argument

from datetime import date as dateType
from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.historical_eps import (
    HistoricalEpsData,
    HistoricalEpsQueryParams,
)
from openbb_core.provider.utils.descriptions import QUERY_DESCRIPTIONS
from pydantic import Field


class FMPHistoricalEpsQueryParams(HistoricalEpsQueryParams):
    """FMP Historical EPS Query.

    Source: https://site.financialmodelingprep.com/developer/docs#earnings-company
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}

    limit: int | None = Field(
        default=None,
        description=QUERY_DESCRIPTIONS.get("limit", "") + " Default is all.",
    )


class FMPHistoricalEpsData(HistoricalEpsData):
    """FMP Historical EPS Data."""

    __alias_dict__ = {
        "eps_actual": "epsActual",
        "eps_estimated": "epsEstimated",
        "revenue_estimated": "revenueEstimated",
        "revenue_actual": "revenueActual",
        "updated": "lastUpdated",
    }

    revenue_estimated: int | float | None = Field(
        default=None,
        description="Estimated consensus revenue for the reporting period.",
    )
    revenue_actual: int | float | None = Field(
        default=None,
        description="The actual reported revenue.",
    )
    updated: dateType | None = Field(
        default=None,
        description="The date when the data was last updated.",
    )


class FMPHistoricalEpsFetcher(
    Fetcher[
        FMPHistoricalEpsQueryParams,
        list[FMPHistoricalEpsData],
    ]
):
    """FMP Historical EPS Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPHistoricalEpsQueryParams:
        """Transform the query params."""
        return FMPHistoricalEpsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPHistoricalEpsQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import warnings  # noqa
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        limit = query.limit + 5 if query.limit is not None else 1000
        results: list = []
        symbols = query.symbol.split(",")

        for symbol in symbols:
            url = f"https://financialmodelingprep.com/stable/earnings?symbol={symbol}&limit={limit}&apikey={api_key}"
            result: list = await get_data_many(url, **kwargs)

            if not result:
                warnings.warn(f"No data found for symbol: {symbol}")
                continue

            results.extend(
                [
                    d
                    for d in sorted(result, key=lambda x: x.get("date"), reverse=True)
                    if d.get("epsActual")
                    or d.get("revenueActual")
                    and d.get("date") <= str(dateType.today())
                ][: query.limit if query.limit is not None else None]
            )

        return results

    @staticmethod
    def transform_data(
        query: FMPHistoricalEpsQueryParams, data: list, **kwargs: Any
    ) -> list[FMPHistoricalEpsData]:
        """Return the transformed data."""
        return [FMPHistoricalEpsData.model_validate(d) for d in data]



"""FMP Earnings Calendar Model."""

# pylint: disable=unused-argument

from datetime import (
    datetime,
    timedelta,
)
from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.calendar_ipo import (
    CalendarIpoData,
    CalendarIpoQueryParams,
)
from pydantic import Field


class FMPCalendarIpoQueryParams(CalendarIpoQueryParams):
    """FMP IPO Calendar Query.

    Source: https://site.financialmodelingprep.com/developer/docs#ipos-calendar
    """

    __alias_dict__ = {
        "start_date": "from",
        "end_date": "to",
    }


class FMPCalendarIpoData(CalendarIpoData):
    """FMP Earnings Calendar Data."""

    __alias_dict__ = {
        "ipo_date": "date",
        "exchange_date": "daa",
        "name": "company",
    }

    exchange_date: datetime | None = Field(
        default=None,
        description="Timezone information for the exchange and date of the IPO.",
    )
    name: str | None = Field(
        default=None,
        description="The name of the entity going public.",
    )
    exchange: str | None = Field(
        default=None,
        description="The exchange where the IPO is listed.",
    )
    actions: str | None = Field(
        default=None,
        description="Actions related to the IPO, such as, Expected, Priced, Filed, Amended.",
    )
    shares: int | float | None = Field(
        default=None,
        description="The number of shares being offered in the IPO.",
    )
    price_range: str | None = Field(
        default=None,
        description="The expected price range for the IPO shares.",
    )
    market_cap: int | float | None = Field(
        default=None,
        description="The estimated market capitalization of the company at the time of the IPO.",
    )


class FMPCalendarIpoFetcher(
    Fetcher[
        FMPCalendarIpoQueryParams,
        list[FMPCalendarIpoData],
    ]
):
    """FMP IPO Calendar Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPCalendarIpoQueryParams:
        """FMP IPO Calendar Query Params."""
        now = datetime.today().date()
        transformed_params = params

        if params.get("start_date") is None:
            transformed_params["start_date"] = now

        if params.get("end_date") is None:
            transformed_params["end_date"] = now + timedelta(days=3)

        return FMPCalendarIpoQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPCalendarIpoQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_core.provider.utils.helpers import amake_requests  # noqa
        from openbb_fmp.utils.helpers import response_callback

        api_key = credentials.get("fmp_api_key") if credentials else ""

        base_url = "https://financialmodelingprep.com/stable/ipos-calendar?"
        start_date = query.start_date or datetime.now().date() - timedelta(days=7)
        end_date = query.end_date or datetime.now().date() + timedelta(days=7)

        # Create 90-day chunks between start_date and end_date
        urls: list = []
        current_start = start_date

        while current_start <= end_date:
            chunk_end = min(current_start + timedelta(days=89), end_date)
            url = f"{base_url}from={current_start}&to={chunk_end}&apikey={api_key}"
            urls.append(url)
            current_start = chunk_end + timedelta(days=1)

        # Get data from all URLs
        all_data: list = await amake_requests(
            urls, response_callback=response_callback, **kwargs
        )

        return all_data

    @staticmethod
    def transform_data(
        query: FMPCalendarIpoQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPCalendarIpoData]:
        """Return the transformed data."""
        return [
            FMPCalendarIpoData.model_validate(d)
            for d in sorted(data, key=lambda x: x["date"], reverse=True)
        ]



"""FMP Yield Curve Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.yield_curve import (
    YieldCurveData,
    YieldCurveQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError

maturity_dict = {
    "month1": "month_1",
    "month2": "month_2",
    "month3": "month_3",
    "month6": "month_6",
    "year1": "year_1",
    "year2": "year_2",
    "year3": "year_3",
    "year5": "year_5",
    "year7": "year_7",
    "year10": "year_10",
    "year20": "year_20",
    "year30": "year_30",
}


class FMPYieldCurveQueryParams(YieldCurveQueryParams):
    """FMP Yield Curve Query.

    Source: https://site.financialmodelingprep.com/developer/docs/treasury-rates-api/
    """

    __json_schema_extra__ = {"date": {"multiple_items_allowed": True}}


class FMPYieldCurveData(YieldCurveData):
    """FMP Yield Curve Data."""


class FMPYieldCurveFetcher(
    Fetcher[
        FMPYieldCurveQueryParams,
        list[FMPYieldCurveData],
    ]
):
    """FMP Yield Curve Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPYieldCurveQueryParams:
        """Transform the query params."""
        transformed_params = params
        if not transformed_params.get("date"):
            transformed_params["date"] = datetime.now().strftime("%Y-%m-%d")
        return FMPYieldCurveQueryParams(**transformed_params)

    @staticmethod
    async def aextract_data(
        query: FMPYieldCurveQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from datetime import timedelta  # noqa
        from openbb_fmp.utils.helpers import get_data_urls

        api_key = credentials.get("fmp_api_key") if credentials else ""

        def generate_url(date):
            """Generate URL for a small window between each date."""
            date = datetime.strptime(date, "%Y-%m-%d")
            from_date = (date - timedelta(days=3)).strftime("%Y-%m-%d")
            to_date = (date + timedelta(days=3)).strftime("%Y-%m-%d")
            url = (
                "https://financialmodelingprep.com/stable/treasury-rates?"
                + f"from={from_date}&to={to_date}&apikey={api_key}"
            )
            return url

        dates = query.date.split(",")  # type: ignore
        urls = [generate_url(date) for date in dates]

        return await get_data_urls(urls, **kwargs)  # type: ignore

    @staticmethod
    def transform_data(
        query: FMPYieldCurveQueryParams, data: list, **kwargs: Any
    ) -> list[FMPYieldCurveData]:
        """Return the transformed data."""
        # pylint: disable=import-outside-toplevel
        from numpy import nan
        from pandas import Categorical, DataFrame, DatetimeIndex

        if not data:
            raise EmptyDataError("The request was returned empty.")
        df = DataFrame(data).set_index("date").sort_index()
        dates = query.date.split(",") if query.date else [df.index.max()]  # type: ignore
        df.index = DatetimeIndex(df.index)
        dates_list = DatetimeIndex(dates)
        df = df.rename(columns=maturity_dict)
        df.columns.name = "maturity"

        # Find the nearest date in the DataFrame to each date in dates_list
        nearest_dates = [df.index.asof(date) for date in dates_list]

        # Filter for only the nearest dates
        df = df[df.index.isin(nearest_dates)]

        df = df.replace({nan: None})

        # Flatten the DataFrame
        flattened_data = df.reset_index().melt(
            id_vars="date", var_name="maturity", value_name="rate"
        )
        flattened_data = flattened_data.sort_values("date")
        flattened_data["maturity"] = Categorical(
            flattened_data["maturity"],
            categories=list(maturity_dict.values()),
            ordered=True,
        )
        flattened_data["rate"] = flattened_data["rate"].astype(float) / 100
        flattened_data = flattened_data.sort_values(
            by=["date", "maturity"]
        ).reset_index(drop=True)
        flattened_data["date"] = flattened_data["date"].dt.strftime("%Y-%m-%d")
        records = flattened_data.to_dict(orient="records")

        return [FMPYieldCurveData.model_validate(d) for d in records]



"""FMP Provider Modules."""

from openbb_core.provider.abstract.provider import Provider
from openbb_fmp.models.analyst_estimates import FMPAnalystEstimatesFetcher
from openbb_fmp.models.available_indices import FMPAvailableIndicesFetcher
from openbb_fmp.models.balance_sheet import FMPBalanceSheetFetcher
from openbb_fmp.models.balance_sheet_growth import FMPBalanceSheetGrowthFetcher
from openbb_fmp.models.calendar_dividend import FMPCalendarDividendFetcher
from openbb_fmp.models.calendar_earnings import FMPCalendarEarningsFetcher
from openbb_fmp.models.calendar_events import FMPCalendarEventsFetcher
from openbb_fmp.models.calendar_ipo import FMPCalendarIpoFetcher
from openbb_fmp.models.calendar_splits import FMPCalendarSplitsFetcher
from openbb_fmp.models.cash_flow import FMPCashFlowStatementFetcher
from openbb_fmp.models.cash_flow_growth import FMPCashFlowStatementGrowthFetcher
from openbb_fmp.models.company_filings import FMPCompanyFilingsFetcher
from openbb_fmp.models.company_news import FMPCompanyNewsFetcher
from openbb_fmp.models.crypto_historical import FMPCryptoHistoricalFetcher
from openbb_fmp.models.crypto_search import FMPCryptoSearchFetcher
from openbb_fmp.models.currency_historical import FMPCurrencyHistoricalFetcher
from openbb_fmp.models.currency_pairs import FMPCurrencyPairsFetcher
from openbb_fmp.models.currency_snapshots import FMPCurrencySnapshotsFetcher
from openbb_fmp.models.discovery_filings import FMPDiscoveryFilingsFetcher
from openbb_fmp.models.earnings_call_transcript import FMPEarningsCallTranscriptFetcher
from openbb_fmp.models.economic_calendar import FMPEconomicCalendarFetcher
from openbb_fmp.models.equity_gainers import FMPGainersFetcher
from openbb_fmp.models.equity_historical import FMPEquityHistoricalFetcher
from openbb_fmp.models.equity_losers import FMPLosersFetcher
from openbb_fmp.models.equity_most_active import FMPEquityActiveFetcher
from openbb_fmp.models.equity_ownership import FMPEquityOwnershipFetcher
from openbb_fmp.models.equity_peers import FMPEquityPeersFetcher
from openbb_fmp.models.equity_profile import FMPEquityProfileFetcher
from openbb_fmp.models.equity_quote import FMPEquityQuoteFetcher
from openbb_fmp.models.equity_screener import FMPEquityScreenerFetcher
from openbb_fmp.models.esg_score import FMPEsgScoreFetcher
from openbb_fmp.models.etf_countries import FMPEtfCountriesFetcher
from openbb_fmp.models.etf_equity_exposure import FMPEtfEquityExposureFetcher
from openbb_fmp.models.etf_holdings import FMPEtfHoldingsFetcher
from openbb_fmp.models.etf_info import FMPEtfInfoFetcher
from openbb_fmp.models.etf_search import FMPEtfSearchFetcher
from openbb_fmp.models.etf_sectors import FMPEtfSectorsFetcher
from openbb_fmp.models.executive_compensation import FMPExecutiveCompensationFetcher
from openbb_fmp.models.financial_ratios import FMPFinancialRatiosFetcher
from openbb_fmp.models.forward_ebitda_estimates import FMPForwardEbitdaEstimatesFetcher
from openbb_fmp.models.forward_eps_estimates import FMPForwardEpsEstimatesFetcher
from openbb_fmp.models.government_trades import FMPGovernmentTradesFetcher
from openbb_fmp.models.historical_dividends import FMPHistoricalDividendsFetcher
from openbb_fmp.models.historical_employees import FMPHistoricalEmployeesFetcher
from openbb_fmp.models.historical_eps import FMPHistoricalEpsFetcher
from openbb_fmp.models.historical_market_cap import FmpHistoricalMarketCapFetcher
from openbb_fmp.models.historical_splits import FMPHistoricalSplitsFetcher
from openbb_fmp.models.income_statement import FMPIncomeStatementFetcher
from openbb_fmp.models.income_statement_growth import FMPIncomeStatementGrowthFetcher
from openbb_fmp.models.index_constituents import FMPIndexConstituentsFetcher
from openbb_fmp.models.index_historical import FMPIndexHistoricalFetcher
from openbb_fmp.models.insider_trading import FMPInsiderTradingFetcher
from openbb_fmp.models.institutional_ownership import FMPInstitutionalOwnershipFetcher
from openbb_fmp.models.key_executives import FMPKeyExecutivesFetcher
from openbb_fmp.models.key_metrics import FMPKeyMetricsFetcher
from openbb_fmp.models.market_snapshots import FMPMarketSnapshotsFetcher
from openbb_fmp.models.nport_disclosure import FMPNportDisclosureFetcher
from openbb_fmp.models.price_performance import FMPPricePerformanceFetcher
from openbb_fmp.models.price_target import FMPPriceTargetFetcher
from openbb_fmp.models.price_target_consensus import FMPPriceTargetConsensusFetcher
from openbb_fmp.models.revenue_business_line import FMPRevenueBusinessLineFetcher
from openbb_fmp.models.revenue_geographic import FMPRevenueGeographicFetcher
from openbb_fmp.models.risk_premium import FMPRiskPremiumFetcher
from openbb_fmp.models.share_statistics import FMPShareStatisticsFetcher
from openbb_fmp.models.treasury_rates import FMPTreasuryRatesFetcher
from openbb_fmp.models.world_news import FMPWorldNewsFetcher
from openbb_fmp.models.yield_curve import FMPYieldCurveFetcher

fmp_provider = Provider(
    name="fmp",
    website="https://financialmodelingprep.com",
    description="""Financial Modeling Prep is a new concept that informs you about
stock market information (news, currencies, and stock prices).""",
    credentials=["api_key"],
    fetcher_dict={
        "AnalystEstimates": FMPAnalystEstimatesFetcher,
        "AvailableIndices": FMPAvailableIndicesFetcher,
        "BalanceSheet": FMPBalanceSheetFetcher,
        "BalanceSheetGrowth": FMPBalanceSheetGrowthFetcher,
        "CalendarDividend": FMPCalendarDividendFetcher,
        "CalendarEarnings": FMPCalendarEarningsFetcher,
        "CalendarEvents": FMPCalendarEventsFetcher,
        "CalendarIpo": FMPCalendarIpoFetcher,
        "CalendarSplits": FMPCalendarSplitsFetcher,
        "CashFlowStatement": FMPCashFlowStatementFetcher,
        "CashFlowStatementGrowth": FMPCashFlowStatementGrowthFetcher,
        "CompanyFilings": FMPCompanyFilingsFetcher,
        "CompanyNews": FMPCompanyNewsFetcher,
        "CryptoHistorical": FMPCryptoHistoricalFetcher,
        "CryptoSearch": FMPCryptoSearchFetcher,
        "CurrencyHistorical": FMPCurrencyHistoricalFetcher,
        "CurrencyPairs": FMPCurrencyPairsFetcher,
        "CurrencySnapshots": FMPCurrencySnapshotsFetcher,
        "DiscoveryFilings": FMPDiscoveryFilingsFetcher,
        "EarningsCallTranscript": FMPEarningsCallTranscriptFetcher,
        "EconomicCalendar": FMPEconomicCalendarFetcher,
        "EquityActive": FMPEquityActiveFetcher,
        "EquityHistorical": FMPEquityHistoricalFetcher,
        "EquityOwnership": FMPEquityOwnershipFetcher,
        "EquityPeers": FMPEquityPeersFetcher,
        "EquityInfo": FMPEquityProfileFetcher,
        "EquityGainers": FMPGainersFetcher,
        "EquityLosers": FMPLosersFetcher,
        "EquityQuote": FMPEquityQuoteFetcher,
        "EquityScreener": FMPEquityScreenerFetcher,
        "EsgScore": FMPEsgScoreFetcher,
        "EtfCountries": FMPEtfCountriesFetcher,
        "EtfEquityExposure": FMPEtfEquityExposureFetcher,
        "EtfHoldings": FMPEtfHoldingsFetcher,
        "EtfInfo": FMPEtfInfoFetcher,
        "EtfPricePerformance": FMPPricePerformanceFetcher,
        "EtfSearch": FMPEtfSearchFetcher,
        "EtfSectors": FMPEtfSectorsFetcher,
        "ExecutiveCompensation": FMPExecutiveCompensationFetcher,
        "FinancialRatios": FMPFinancialRatiosFetcher,
        "ForwardEbitdaEstimates": FMPForwardEbitdaEstimatesFetcher,
        "ForwardEpsEstimates": FMPForwardEpsEstimatesFetcher,
        "HistoricalDividends": FMPHistoricalDividendsFetcher,
        "HistoricalEmployees": FMPHistoricalEmployeesFetcher,
        "HistoricalEps": FMPHistoricalEpsFetcher,
        "HistoricalMarketCap": FmpHistoricalMarketCapFetcher,
        "HistoricalSplits": FMPHistoricalSplitsFetcher,
        "IncomeStatement": FMPIncomeStatementFetcher,
        "IncomeStatementGrowth": FMPIncomeStatementGrowthFetcher,
        "IndexConstituents": FMPIndexConstituentsFetcher,
        "IndexHistorical": FMPIndexHistoricalFetcher,
        "InsiderTrading": FMPInsiderTradingFetcher,
        "InstitutionalOwnership": FMPInstitutionalOwnershipFetcher,
        "KeyExecutives": FMPKeyExecutivesFetcher,
        "KeyMetrics": FMPKeyMetricsFetcher,
        "MarketSnapshots": FMPMarketSnapshotsFetcher,
        "NportDisclosure": FMPNportDisclosureFetcher,
        "PricePerformance": FMPPricePerformanceFetcher,
        "PriceTarget": FMPPriceTargetFetcher,
        "PriceTargetConsensus": FMPPriceTargetConsensusFetcher,
        "RevenueBusinessLine": FMPRevenueBusinessLineFetcher,
        "RevenueGeographic": FMPRevenueGeographicFetcher,
        "RiskPremium": FMPRiskPremiumFetcher,
        "ShareStatistics": FMPShareStatisticsFetcher,
        "TreasuryRates": FMPTreasuryRatesFetcher,
        "WorldNews": FMPWorldNewsFetcher,
        "EtfHistorical": FMPEquityHistoricalFetcher,
        "YieldCurve": FMPYieldCurveFetcher,
        "GovernmentTrades": FMPGovernmentTradesFetcher,
    },
    repr_name="Financial Modeling Prep (FMP)",
    deprecated_credentials={"API_KEY_FINANCIALMODELINGPREP": "fmp_api_key"},
    instructions='Go to: https://site.financialmodelingprep.com/developer/docs\n\n![FinancialModelingPrep](https://user-images.githubusercontent.com/46355364/207821920-64553d05-d461-4984-b0fe-be0368c71186.png)\n\nClick on, "Get my API KEY here", and sign up for a free account.\n\n![FinancialModelingPrep](https://user-images.githubusercontent.com/46355364/207822184-a723092e-ef42-4f87-8c55-db150f09741b.png)\n\nWith an account created, sign in and navigate to the Dashboard, which shows the assigned token. by pressing the "Dashboard" button which will show the API key.\n\n![FinancialModelingPrep](https://user-images.githubusercontent.com/46355364/207823170-dd8191db-e125-44e5-b4f3-2df0e115c91d.png)',  # noqa: E501  pylint: disable=line-too-long
)



"""FMP Cash Flow Statement Growth Model."""

# pylint: disable=unused-argument

from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.cash_flow_growth import (
    CashFlowStatementGrowthData,
    CashFlowStatementGrowthQueryParams,
)
from openbb_core.provider.utils.descriptions import (
    DATA_DESCRIPTIONS,
    QUERY_DESCRIPTIONS,
)
from openbb_fmp.utils.definitions import FinancialPeriods
from pydantic import Field


class FMPCashFlowStatementGrowthQueryParams(CashFlowStatementGrowthQueryParams):
    """FMP Cash Flow Statement Growth Query.

    Source: https://site.financialmodelingprep.com/developer/docs#cashflow-statement-growth
    """

    period: FinancialPeriods = Field(
        default="annual",
        description=QUERY_DESCRIPTIONS.get("period", ""),
    )


class FMPCashFlowStatementGrowthData(CashFlowStatementGrowthData):
    """FMP Cash Flow Statement Growth Data."""

    __alias_dict__ = {
        "period_ending": "date",
        "fiscal_year": "calendarYear",
        "fiscal_period": "period",
        "reported_currency": "reportedCurrency",
        "growth_acquisitions": "growthAcquisitionsNet",
        "growth_sale_and_maturity_of_investments": "growthSalesMaturitiesOfInvestments",
        "growth_net_cash_from_operating_activities": "growthNetCashProvidedByOperatingActivities",
        "growth_other_investing_activities": "growthOtherInvestingActivities",
        "growth_net_cash_from_investing_activities": "growthNetCashUsedForInvestingActivities",
        "growth_other_financing_activities": "growthOtherFinancingActivities",
        "growth_purchase_of_investment_securities": "growthPurchasesOfInvestments",
        "growth_account_receivables": "growthAccountsReceivables",
        "growth_account_payable": "growthAccountsPayables",
        "growth_purchase_of_property_plant_and_equipment": "growthInvestmentsInPropertyPlantAndEquipment",
        "growth_repayment_of_debt": "growthDebtRepayment",
        "growth_net_change_in_cash_and_equivalents": "growthNetChangeInCash",
        "growth_effect_of_exchange_rate_changes_on_cash": "growthEffectOfForexChangesOnCash",
        "growth_net_cash_from_financing_activities": "growthNetCashUsedProvidedByFinancingActivities",
        "growth_net_equity_issuance": "growthNetStockIssuance",
        "growth_common_equity_issuance": "growthCommonStockIssued",
        "growth_common_equity_repurchased": "growthCommonStockRepurchased",
    }
    symbol: str = Field(description=DATA_DESCRIPTIONS.get("symbol", ""))
    reported_currency: str | None = Field(
        description="The currency in which the financial data is reported.",
        default=None,
    )
    growth_net_income: float | None = Field(
        default=None,
        description="Growth rate of net income.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_depreciation_and_amortization: float | None = Field(
        default=None,
        description="Growth rate of depreciation and amortization.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_deferred_income_tax: float | None = Field(
        default=None,
        description="Growth rate of deferred income tax.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_stock_based_compensation: float | None = Field(
        default=None,
        description="Growth rate of stock-based compensation.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_change_in_working_capital: float | None = Field(
        default=None,
        description="Growth rate of change in working capital.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_account_receivables: float | None = Field(
        default=None,
        description="Growth rate of accounts receivables.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_inventory: float | None = Field(
        default=None,
        description="Growth rate of inventory.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_account_payable: float | None = Field(
        default=None,
        description="Growth rate of account payable.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_other_working_capital: float | None = Field(
        default=None,
        description="Growth rate of other working capital.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_other_non_cash_items: float | None = Field(
        default=None,
        description="Growth rate of other non-cash items.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_net_cash_from_operating_activities: float | None = Field(
        default=None,
        description="Growth rate of net cash provided by operating activities.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_purchase_of_property_plant_and_equipment: float | None = Field(
        default=None,
        description="Growth rate of investments in property, plant, and equipment.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_acquisitions: float | None = Field(
        default=None,
        description="Growth rate of net acquisitions.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_purchase_of_investment_securities: float | None = Field(
        default=None,
        description="Growth rate of purchases of investments.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_sale_and_maturity_of_investments: float | None = Field(
        default=None,
        description="Growth rate of sales maturities of investments.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_other_investing_activities: float | None = Field(
        default=None,
        description="Growth rate of other investing activities.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_net_cash_from_investing_activities: float | None = Field(
        default=None,
        description="Growth rate of net cash used for investing activities.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_short_term_net_debt_issuance: float | None = Field(
        default=None,
        description="Growth rate of short term net debt issuance.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_long_term_net_debt_issuance: float | None = Field(
        default=None,
        description="Growth rate of long term net debt issuance.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_net_debt_issuance: float | None = Field(
        default=None,
        description="Growth rate of net debt issuance.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_repayment_of_debt: float | None = Field(
        default=None,
        description="Growth rate of debt repayment.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_common_equity_issuance: float | None = Field(
        default=None,
        description="Growth rate of common equity issued.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_common_equity_repurchased: float | None = Field(
        default=None,
        description="Growth rate of common equity repurchased.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_net_equity_issuance: float | None = Field(
        default=None,
        description="Growth rate of net equity issuance.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_dividends_paid: float | None = Field(
        default=None,
        description="Growth rate of dividends paid.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_preferred_dividends_paid: float | None = Field(
        default=None,
        description="Growth rate of preferred dividends paid.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_other_financing_activities: float | None = Field(
        default=None,
        description="Growth rate of other financing activities.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_net_cash_from_financing_activities: float | None = Field(
        default=None,
        description="Growth rate of net cash used/provided by financing activities.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_effect_of_exchange_rate_changes_on_cash: float | None = Field(
        default=None,
        description="Growth rate of the effect of foreign exchange changes on cash.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_net_change_in_cash_and_equivalents: float | None = Field(
        default=None,
        description="Growth rate of net change in cash.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_cash_at_beginning_of_period: float | None = Field(
        default=None,
        description="Growth rate of cash at the beginning of the period.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_cash_at_end_of_period: float | None = Field(
        default=None,
        description="Growth rate of cash at the end of the period.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_operating_cash_flow: float | None = Field(
        default=None,
        description="Growth rate of operating cash flow.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_capital_expenditure: float | None = Field(
        default=None,
        description="Growth rate of capital expenditure.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_income_taxes_paid: float | None = Field(
        default=None,
        description="Growth rate of income taxes paid.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_interest_paid: float | None = Field(
        default=None,
        description="Growth rate of interest paid.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    growth_free_cash_flow: float | None = Field(
        default=None,
        description="Growth rate of free cash flow.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )


class FMPCashFlowStatementGrowthFetcher(
    Fetcher[
        FMPCashFlowStatementGrowthQueryParams,
        list[FMPCashFlowStatementGrowthData],
    ]
):
    """FMP Cash Flow Statement Growth Fetcher."""

    @staticmethod
    def transform_query(
        params: dict[str, Any],
    ) -> FMPCashFlowStatementGrowthQueryParams:
        """Transform the query params."""
        return FMPCashFlowStatementGrowthQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPCashFlowStatementGrowthQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list[dict]:
        """Transform the query, extract and transform the data from the FMP endpoints."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""

        url = (
            "https://financialmodelingprep.com/stable/cash-flow-statement-growth"
            + f"?symbol={query.symbol}"
            + f"&period={query.period}"
            + f"&limit={query.limit if query.limit else 5}"
            + f"&apikey={api_key}"
        )

        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPCashFlowStatementGrowthQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPCashFlowStatementGrowthData]:
        """Return the transformed data."""
        return [FMPCashFlowStatementGrowthData.model_validate(d) for d in data]
